from . import jsonFile



def open():
    filename = "/Users/uormaechea/Documents/Dev/python/Industries copy 3/tests/output/debug/11:58:40--getBasketDetail_res.json"
    dic = jsonFile.read(filename)
    return dic

def find(dic, path):
    chunks = path.split(".")
    return _find(dic,chunks)

def _pop0(paths):
    p = paths.copy()
    return p.pop(0)

def _find(dic,paths):
    if type(dic) is list:
        for d in dic:
            _find(dic,paths)

    for key in dic.keys():
        if key == paths[0]:
            _paths = _pop0(paths)
            if len(_paths) == 0:
                return dic[key]
            else:
                found = _find(dic['key'],_paths)
                if found != None:
                    return found
        _obj = dic[key]
        if type(_obj) is list or type(_obj) is dict:
            return 
        if type(_obj) is dict:
            return 2
        return None
    


import unittest,simplejson
from InCli import InCli
from InCli.SFAPI import account,restClient,CPQ,query,file_csv,jsonFile

class Test_Account(unittest.TestCase):
    def test_main(self):
        name = 'User_a2'
        accountF = f"Name:{name}"

        restClient.init('DTI')
        restClient.setLoggingLevel()
        accountId = account.create_Id(name,recordTypeName='Consumer',checkExists=True)

        dele = account.deleteOrders(accountF)

        cartId = CPQ.createCart(accountF,'Name:B2C Price List','name')
        orders = account.getOrdersId(accountF)
        self.assertTrue(orders!=None)

        dele = account.deleteOrders(accountF)
        orders = account.getOrdersId(accountF)
        self.assertTrue(orders==[])

        accountId2 = account.getId(accountF)
        self.assertTrue(accountId==accountId2)

        acc = account.get(accountId)
        self.assertTrue(accountId==acc['Id'])

        account.delete(accountId)
        acc2 = account.get(accountId)
        self.assertTrue(acc2==None)
        accountId3 = account.getId(f"Name:{name}")
        self.assertTrue(accountId3==None)

        acc3 = account.create(name,recordTypeName='Consumer',checkExists=True)
        accountId4 = account.getId(accountF=f"Name:{name}")
        self.assertTrue(accountId4==acc3['Id'])

        acc4 = account.create(name,recordTypeName='Consumer',checkExists=True)
        self.assertTrue(accountId4==acc4['Id'])
        account.delete(accountId4)

        account.delete(accountId4)


        print()

    def test_get_errors(self):
        restClient.init('NOSPRD')

        q = f"select Id,vlocity_cmt__InputData__c from vlocity_cmt__VlocityErrorLogEntry__c"

        res = query.query(q)


        out=[]
        for record in res['records']:
            data = record['vlocity_cmt__InputData__c']
            datas = simplejson.loads(data)
            out.append(datas)
            print(datas['ErrorMessage__c'])


        file_csv.write('VlocityErrorLogEntry_1',out)

        print()

    def test_get_error_ip(self):
        restClient.init('NOSDEV')

        Id = 'a6K3O000000FHuqUAG'

        q = f"select fields(all) from vlocity_cmt__VlocityErrorLogEntry__c where Id='{Id}' limit 10"

        res = query.query(q)


        out=[]
        for record in res['records']:
            data = record['vlocity_cmt__InputData__c']
            datas = simplejson.loads(data)

            theFile =jsonFile.write('TheIPError_1',datas)
            
            out.append(datas)
            print(datas['ErrorMessage__c'])


      #  file_csv.write('VlocityErrorLogEntry_1',out)

        print()

        



from . import Sobjects,tooling,utils
#def get_traceFlag_for_user(userF):
#    userId = Sobjects.IdF('User',userF)
#    q = f"select id, TracedEntityId,logtype, startdate, expirationdate, debuglevelid, debuglevel.apexcode, debuglevel.visualforce from TraceFlag where TracedEntityId = '{userId}'"
#    call = query(q)
#    utils.printFormated(call['records'],fieldsString="Id StartDate ExpirationDate DebugLevel.ApexCode",rename="DebugLevel.ApexCode%ApexCode",separator=' ')

def create_debug_level_incli():
    data = {
        "DeveloperName": "InCli",
        "Language": "pt_BR",
        "MasterLabel": "InCli",
        "Workflow": "FINE",
        "Validation": "INFO",
        "Callout": "INFO",
        "ApexCode": "FINEST",
        "ApexProfiling": "INFO",
        "Visualforce": "FINER",
        "System": "FINE",
        "Database": "FINEST",
        "Wave": "NONE",
        "Nba": "NONE"
    }
    call = tooling.post('DebugLevel',data=data)
    return call


def create_trace_flag_incli(userId,DebugLevelId):
    data = {
        "TracedEntityId":userId,
        "LogType":"USER_DEBUG",
        "DebugLevelId":DebugLevelId,
        "StartDate":utils.datetime_now_string(),
        "ExpirationDate":utils.datetime_now_string(addMinutes=10)
    }
    call = tooling.post('TraceFlag',data=data)
    return call

    #'7tf3O000001LbmoQAC'

def update_trace_flag_incli(id,minutes=5,start=-2):
    data = {
        "StartDate":utils.datetime_now_string(addMinutes=start),
        "ExpirationDate":utils.datetime_now_string(addMinutes=minutes)
    }
    call = tooling.patch(sobject='TraceFlag',id=id , data=data)
    tooling.checkError()
    return call

def get_traceflags_for_user(userF):
    userId = Sobjects.IdF('User',userF)

    q = f"select Id,StartDate,ExpirationDate,DebugLevelId from TraceFlag where TracedEntityId='{userId}'"
    call = tooling.query(q)
    print()

def set_incli_traceFlag_for_user(userF):
    userId = Sobjects.IdF('User',userF)
    DebugLevelId = tooling.idF('DebugLevel',"DeveloperName:InCli")   
    if DebugLevelId == None:
        call = create_debug_level_incli()
        DebugLevelId = call['id']  

    q = f"select Id from TraceFlag where DebugLevelId='{DebugLevelId}' and TracedEntityId = '{userId}'"
    call = tooling.query(q)
 #   traceFlagId = tooling.idF('TraceFlag',f"DebugLevelId:{DebugLevelId}")
  #  tf = tooling.get('TraceFlag',traceFlagId)
    if len(call['records']) == 0:
        call = create_trace_flag_incli(userId,DebugLevelId)
        traceFlagId = call['id']
    else:
       traceFlagId = call['records'][0]['Id']

    update_trace_flag_incli(traceFlagId,5)

    return traceFlagId
    
#	https://appsmobileqms.nos.pt
#   https://appsmobileqms.nos.pt/
#   https://appsmobileqms.nos.pt


import logging,json,simplejson
#from posixpath import split
from . import restClient as client
from . import Sobjects,utils,thread
" select fields(all) from vlocity_cmt__EntityFilter__c where Id in ('a5W7Z0000009gEiUAI','a5W7Z0000009gdtUAA','a5W7Z0000009gFfUAI','a5W7Z0000009gEdUAI')"

def query_base64(q,raiseEx=True):
    call = client.callAPI(f'/services/data/v55.0/query?q={q}')
    checkError(q,raiseEx)
    if call['totalSize'] > len(call['records']):
        len_max_records = len(call['records'])
        q_order = f"{q} order by Id asc"
        call = _query(q_order,nextRecords=False)
        len_records_last_call = len(call['records'])
        call_next = call
        while (len_records_last_call==len_max_records):
            lastId = call_next['records'][len_max_records-1]['Id']
            q_and= f"{q} and Id>'{lastId}' order by Id asc"
            call_next = _query(q_and,nextRecords=False)
            call['records'].extend(call_next['records'])
            len_records_last_call = len(call_next['records'])

    return call

def query_threaded(q,values,search="$$$",raiseEx=True):
    result = []

    def do_work(value):
        q1 = q.replace(search,value)
        res = _query(q1)
        return res

    def on_done(res,result):
        result.append(res)

    thread.execute_threaded(values,result,do_work,on_done,threads=1)

def checkError(q,raiseEx=True):
    lc = client.thread_lastCall()

    lc['query']=q
    if lc['error'] != None:
        lc['query'] = q
        if raiseEx==True:
            errorOther = lc['errorOther'] if 'errorOther' in lc else ''
            utils.raiseException(lc['errorCode'],lc['error'],'query',f"{errorOther} . {q}")
        return True    
    return False

def _query(q,raiseEx=True,logResponse = False,nextRecords=True,popAttributes=True):
    logging.debug(q)
    call = client.callAPI(f'/services/data/v55.0/query?q={q}')
   # print(len(call['records']))

    call_next = call
    if nextRecords:
        while 'nextRecordsUrl' in call_next:
            call_next = client.callAPI(call_next['nextRecordsUrl'])
          #  print(len(call_next['records']))
            call['records'].extend(call_next['records'])

    if checkError(q,raiseEx) == True:
        logging.warn(simplejson.dumps(call, indent=4))
        return None
  #  lc = client.thread_lastCall()
  #  lc['query']=q
  #  if lc['error'] != None:
  #      lc['query'] = q
  #      if raiseEx==True:
  #          errorOther = lc['errorOther'] if 'errorOther' in lc else ''
  #          utils.raiseException(lc['errorCode'],lc['error'],'query',f"{errorOther} . {q}")
  #          raise ValueError(simplejson.dumps(call, indent=4))
  #      logging.warn(simplejson.dumps(call, indent=4))

  #      return None

    if logResponse == True:
        utils.printJSON(call)
    for r in call['records']: r.pop('attributes')

    return call

#--------------------------------------------------------------------------------------------------------------------------------------------------------
def query(q,raiseEx=True,X=['Id','Name'],logResponse=False):
    """
    Executes a query in Salesforce.
    - q: the query string. "select.... from... limit..."
        the where clause can specify a normal field or $X and $Z. 
    - X: by default it will query for Id and Name. 
    - logResponse: 
    """
    if '$X=' not in q and '$Z=' not in q:
        return _query(q,raiseEx,logResponse)

    if '$Z='  in q:
        X=['Id','ProductCode','Name']
        q = q.replace('$Z=','$X=')

    selectfields = q.strip().split()

    for x in range(len(selectfields)):
        if selectfields[x].lower() == 'from':
            objectType = selectfields[x+1]

        if  '$X' in selectfields[x]:
            where = selectfields[x].split("'")
            value = where[1]

    for field in X:
        if field == 'Id':
            if Sobjects.isId(objectType,value) == False:
                continue
        q1 = q.replace('$X',field)
        call = _query(q1,raiseEx=False)      
        if call['totalSize'] > 0:
            return call

    return None

def queryRecords(q,raiseEx=True,X=['Id','Name']):
    select = query(q,raiseEx,X)
    if select == None:
        return None
    return select['records']

cache = {}
def queryFieldList(q,field=None,raiseEx=False,X=['Id','Name']):
    if '$X=' in q or '$Z=' in q:
        print(q)
        if q in cache:
            return cache[q]
    records = queryRecords(q,raiseEx,X)

    if records == None:
        return None

    if field == None:
        field = q.strip().split()[1]
    
    fieldList = []
  
    for record in records:
        fieldList.append(record[field])
        
    if '$X=' in q or '$Z=' in q:
        cache[q]=fieldList

    return fieldList

#" select Id from vlocity_cmt__AttributeAssignment__c where vlocity_cmt__AttributeUniqueCode__c='ATT_MOBILE_CREDITS' and  vlocity_cmt__ObjectId__c='01t7Z00000AVpCLQA1' "
def queryField(q,field=None,raiseEx=False,X=['Id','Name']):
    fieldList = queryFieldList(q,field,raiseEx,X)

    if fieldList == None or len(fieldList) == 0:
        return None

    if len(fieldList) > 1:
        logging.warn(f"There is more than one record returned. total records {len(fieldList)}, query={q}")

    return fieldList[0]

def queryIdF(objName,extendedF,init=None):
    """Returns the Id for the especified extendedF.
    - if extendedF is a string, returns the string
    - if extendedF is a dictionary, return extendedF['Id]
    - if extendedF is a fieldName:fieldValue, returns the query on object where fieldName=fieldValue
    """
    if extendedF == None:
        return None
    if type(extendedF) is dict:
        extendedF = extendedF['Id']
    ef = utils.extendedField(extendedF)
    if ef['field'] == 'Id':
        if Sobjects.checkId(ef['value']):
            return ef['value']
        else:
            utils.raiseException("No_Id",f"{extendedF} is not a valid Id")
    return queryField(f" select Id from {objName} where {ef['field']} = '{ef['value']}' ")
def logQuery(q,filename,raiseEx=False):
    if filename != None:
        client.logCall(filename)
    call = query(q,raiseEx)

    return call


def process( call):
    for asset in call['records']:
        asset['vlocity_cmt__PricingLogData__c'] = json.loads( asset['vlocity_cmt__PricingLogData__c'])

def take_snapshot(fileName,q,postProcessing=None):
    client.debugFile(fileName)
    client.setDebugReplyProcessing(postProcessing)
    query(q)

def records(call):
    return call['records']

def  IN_clause(list):
    values = []
    for p in list:
        values.append(f"'{p}'")
    return ",".join(values)


import unittest
from InCli.SFAPI import restClient,CPQ,account,Sobjects,utils,query,jsonFile,objectUtil,file,timeStats,digitalCommerce,digitalCommerceUtil
import simplejson,threading,time,multiprocessing,traceback,random,uuid,time,calendar
import InCli.InCli as incli

procNum = 10
threadNum = 20
reps = 30
guest_url = 'https://nosdti-parceiros.cs109.force.com/onboarding'
def do_work(details):
  restClient.initWithToken('GUEST',url=guest_url)

  def getOffers(details):

    errors = 0
    ts = timeStats.TimeStats()

    for x in range(0,reps):
      try:
        # cartId = '8017a000002ad21AAA'
        # items2 = CPQ.getCartItems_api(cartId,price=False,validate=False)

        #
        # call = digitalCommerce.getOfferByCatalogue('DC_CAT_WOO_MOBILE')
        # digitalCommerce.createBasket('DC_CAT_WOO_MOBILE','C_WOO_MOBILE')
      #  val = f"{uuid.uuid4()}"
      #  print(val)
        ts.new(["Basket","Error"])
        current_GMT = time.gmtime()

        val = calendar.timegm(current_GMT)
        offerDetails = digitalCommerceUtil.updateOfferField(details,'ATT_NOS_OTT_SUBSCRIPTION_ID','userValues',val,'code')
        print("creating basket")        
        basket2 = digitalCommerce.createBasketAfterConfig('DC_CAT_WOO_MOBILE',offerDetails)
        print("got basket")
        ts.time("Basket")

        call1 = restClient.lastCall()
        if call1['totalCalls'] >1:
          calls = restClient.thread_get_calls()
          times = restClient.get_call_times(call1)
          times2 = restClient.get_call_times(calls[-2])
          ts.time_no("Basket1",times2['delta'])
          ts.time_no("Basket2",times['delta'])

       #   print(f"{times2} {times}",flush=True)

     #   print(restClient.getLastCallTimes(),flush=True)
      except Exception as e:
        utils.printException(e)
        errors = errors + 1
        ts.time_no("Error","True")

    ts.print()
      
    print(f"Processed {reps} with errors{errors}")  
     #   print(traceback.format_exc())
     #   print(restClient.lastCall())

  threads =  []
  for x in range(0,threadNum):
      thread = threading.Thread(target=getOffers,args=(details,))
      thread.start()
      threads.append(thread)
  for x,t in enumerate(threads):
    t.join()
    print(f'Thread Done {x}')


class Test_Query(unittest.TestCase):
    def test_guestUser(self):
        def do_process():
          restClient.initWithToken('GUEST',url=guest_url)

          print("getting offer details")
          details = digitalCommerce.getOfferDetails('DC_CAT_WOO_MOBILE','C_WOO_MOBILE')
          print("got offer details")

          processes = []
          for x in range(0,procNum):
            process = multiprocessing.Process(target=do_work,args=(details,))
            process.start()  
            processes.append(process)
          for p in processes:
            p.join()
            print('Process Done')

        do_process()

    def test_query_apexlog(self):
      restClient.init('DTI')

      q = "select Operation,StartTime,DurationMilliseconds from Apexlog order by StartTime desc limit 200"
      res = query.query(q)
      for r in res['records']:
        r.pop('attributes')

      utils.printFormated(res['records'])
      print()
    def test_createCart(self):
        restClient.init('NOSDEV')

        acc = account.createAccount('Name:unaiTest2',recordTypeName='Consumer')

        self.assertTrue(Sobjects.checkId(acc['Id']))

        cartId = CPQ.createCart(accountF= acc['Id'], pricelistF='Name:B2C Price List',name="testa5",checkExists=True)

        print(cartId)

    def test_delCartItems(self):
        restClient.init('NOSDEV')
        cartId = CPQ.getCartId('name:testa5')

        items = CPQ.getCartItems_api(cartId)

        if items['totalSize'] > 0:
          itemIds = [item['Id']['value'] for item in items['records']]

          dele = CPQ.deleteCartItems_api(cartId,itemIds)

          self.assertEqual(dele['messages'][0]['message'],'Successfully deleted.')
          print()

    def test_delete_applied_promotions(self):
        restClient.init('NOSDEV')

        cartId = CPQ.getCartId('name:testa5')
        applied = CPQ.getCartPromotions_api(cartId,getPromotionsAppliedToCart=True)
        dele = CPQ.deleteCartPromotion_api(cartId,applied[0]['Id']['value'])

    def test_postPromo(self):
        restClient.init('NOSDEV')

        promoName = 'NOS4u 100Mb + Móvel'
        promo = CPQ.getCartPromotions('name:testa5',query=promoName,onlyOne=True)

        cartId = CPQ.getCartId('name:testa5')

        res = CPQ.postCartsPromoItems_api(cartId,promo['Id'])

        applied = CPQ.getCartPromotions_api(cartId,getPromotionsAppliedToCart=True)

       # dele = CPQ.deleteCartPromotion_api(cartId,applied[0]['Id']['value'])

        print()

    def test_getErrors(self):
        cartId = CPQ.getCartId('name:testa5')

        all = CPQ.getCartItems_api(cartId,fields='vlocity_cmt__ServiceAccountId__c,vlocity_cmt__BillingAccountId__c')

        sim = all['records'][0]['lineItems']['records'][3]['productGroups']['records'][1]['productGroups']['records'][0]['lineItems']['records'][0]
        for key in sim.keys():
          sim2= sim.copy()
          sim2.pop(key)
          items = {"records":[sim2]}
          res = CPQ.updateCartItem_api(cartId,items)
          if res['messages'][0]['message'] == 'Successfully updated.':
            print(f"{key} {res['messages'][0]['message']}")
          else:
            print(f"******************************************************   {key} {res['messages'][0]['message']}")

    def test_updateCartItem(self):
        restClient.init('NOSDEV')
        cartId = CPQ.getCartId('name:testa5')

        all = CPQ.getCartItems_api(cartId,fields='vlocity_cmt__ServiceAccountId__c,vlocity_cmt__BillingAccountId__c')

        sim = all['records'][0]['lineItems']['records'][3]['productGroups']['records'][1]['productGroups']['records'][0]['lineItems']['records'][0]

        print(sim['Quantity'])
        sim_m = {}
        sim_m['vlocity_cmt__LineNumber__c'] = sim['vlocity_cmt__LineNumber__c']
        sim_m['vlocity_cmt__RootItemId__c'] = sim['vlocity_cmt__RootItemId__c']
        sim_m['vlocity_cmt__Action__c'] = sim['vlocity_cmt__Action__c']
        sim_m['Id'] = sim['Id']
      #  sim_m['Quantity'] = sim['Quantity']
      #  sim_m['Quantity']['value'] = 2.0
        items = {"records":[sim_m]}

        sim_m['vlocity_cmt__ServiceAccountId__c'] = '0013O000016sKfwQAE'
        sim_m['vlocity_cmt__BillingAccountId__c'] = '0013O000016sKfwQAE'

        res = CPQ.updateCartItem_api(cartId,items)
        time = restClient.getLastCallTime()
        print(time)
        print(res['messages'][0]['message'])

        all_2 = CPQ.getCartItems_api(cartId,fields='vlocity_cmt__ServiceAccountId__c,vlocity_cmt__BillingAccountId__c')
        sim_2 = all_2['records'][0]['lineItems']['records'][3]['productGroups']['records'][1]['productGroups']['records'][0]['lineItems']['records'][0]
        print(sim_2['Quantity']['value'])
        print()

    def test_updateCartItem(self):
        restClient.init('NOSDEV')

        postPromo =True

        acc = account.createAccount('Name:unaiTest2',recordTypeName='Consumer')
        cartId = CPQ.createCart(accountF= acc['Id'], pricelistF='Name:B2C Price List',name="testa_7",checkExists=True)
        print(cartId)

        if postPromo==True:
          promoName = 'NOS4u 100Mb + Móvel'
          promo = CPQ.getCartPromotions(cartId,query=promoName,onlyOne=True)

          res = CPQ.postCartsPromoItems_api(cartId,promo['Id'])

        all = CPQ.getCartItems_api(cartId,fields='vlocity_cmt__ServiceAccountId__c,vlocity_cmt__BillingAccountId__c')

        codes = ['C_NOS_OFFER_001','C_NOS_AGG_SERVICE_TV_UMA','C_NOS_SERVICE_TV_003','C_NOS_EQUIP_TV_009']
        codes = ['C_NOS_OFFER_001']

        account1Id = acc['Id']
        account2Id = account.getAccountId('name:unai2')

        objs = []
        for code in codes:
          lineItem = objectUtil.getSiblingWhere(all,selectKey='ProductCode',selectKeyValue=code,whereKey='itemType',whereValue='lineItem')['object']
          if lineItem == None:
            print(f"None for code {code}")
            continue

          assert(lineItem['vlocity_cmt__ServiceAccountId__c']['value'] == account1Id)
          assert(lineItem['vlocity_cmt__BillingAccountId__c']['value'] == account1Id)

          obj = {}
          obj['vlocity_cmt__LineNumber__c'] = lineItem['vlocity_cmt__LineNumber__c']
          obj['vlocity_cmt__RootItemId__c'] = lineItem['vlocity_cmt__RootItemId__c']
          obj['vlocity_cmt__Action__c'] = lineItem['vlocity_cmt__Action__c']
          obj['vlocity_cmt__AssetReferenceId__c'] = lineItem['vlocity_cmt__AssetReferenceId__c']
          obj['Id'] = lineItem['Id']
          obj['vlocity_cmt__ServiceAccountId__c'] = account2Id
          obj['vlocity_cmt__BillingAccountId__c'] = account2Id

          objs.append(obj)
       #   objs.append({"records":[obj]})

        items = {"records":objs}
        res = CPQ.updateCartItem_api(cartId,items)
        time = restClient.getLastCallTime()
        restClient.callSave('minimum',logRequest=True)
        print(time)
        print(res['messages'][0]['message'])

#        all2 = CPQ.getCartItems_api(cartId,fields='vlocity_cmt__ServiceAccountId__c,vlocity_cmt__BillingAccountId__c')
#
#        for code in codes:
#          lineItem = objectUtil.getSiblingWhere(all2,selectKey='ProductCode',selectKeyValue=code,whereKey='itemType',whereValue='lineItem')['object']
#          print(f"{code}  {lineItem['vlocity_cmt__ServiceAccountId__c']['value']}  {lineItem['vlocity_cmt__BillingAccountId__c']['value']}   {account2Id}")

        incli._main(["-u","NOSDEV","-q", f"select Id,vlocity_cmt__BillingAccountId__c,vlocity_cmt__LineNumber__c,vlocity_cmt__ServiceAccountId__c,vlocity_cmt__Product2Id__r.productCode from orderitem where OrderId='{cartId}' order by vlocity_cmt__LineNumber__c","-fields","all"])  

        CPQ.deleteCart(cartId)

        print()

    def test_write_to_file(self):
        restClient.init('NOSDEV')

        cartId = CPQ.getCartId('name:testa_1')
        CPQ.deleteCart(cartId)

        acc = account.createAccount('Name:unaiTest2',recordTypeName='Consumer')
        cartId = CPQ.createCart(accountF= acc['Id'], pricelistF='Name:B2C Price List',name="testa_1",checkExists=True)
        print(cartId)

        promoName = 'NOS4u 100Mb + Móvel'
        promo = CPQ.getCartPromotions(cartId,query=promoName,onlyOne=True)

        res = CPQ.postCartsPromoItems_api(cartId,promo['Id'])

        all = CPQ.getCartItems_api(cartId,fields='vlocity_cmt__ServiceAccountId__c,vlocity_cmt__BillingAccountId__c')
        filename = restClient.callSave('getItemsqwe')
        print(filename)

    def test_read_from_file(self):
        restClient.init('NOSDEV')
        account1Id = account.getAccountId('name:unaiTest2')
        account2Id = account.getAccountId('name:unai2')

        print(account2Id)

        stItems = file.read('/Users/uormaechea/Documents/Dev/python/InCliLib/.incli/output/getItemsqwe_res.json')
        stItems2 = stItems.replace(account1Id,account2Id)
        items = simplejson.loads(stItems2)
        cartId = CPQ.getCartId('name:testa_1')
        print(cartId)
     #   res = CPQ.updateCartItem_api(cartId,items)

        C_NOS_OFFER_001 = objectUtil.getSiblingWhere(items,selectKey='ProductCode',selectKeyValue='C_NOS_OFFER_001',whereKey='itemType',whereValue='lineItem')['object']
        C_NOS_EQUIP_TV_006 = objectUtil.getSiblingWhere(items,selectKey='ProductCode',selectKeyValue='C_NOS_EQUIP_TV_006',whereKey='itemType',whereValue='lineItem')['object']
        C_NOS_EQUIP_TV_009 = objectUtil.getSiblingWhere(items,selectKey='ProductCode',selectKeyValue='C_NOS_EQUIP_TV_009',whereKey='itemType',whereValue='lineItem')['object']
        C_NOS_SERVICE_TV_003 = objectUtil.getSiblingWhere(items,selectKey='ProductCode',selectKeyValue='C_NOS_SERVICE_TV_003',whereKey='itemType',whereValue='lineItem')['object']

        C_NOS_SERVICE_IF_001 = objectUtil.getSiblingWhere(items,selectKey='ProductCode',selectKeyValue='C_NOS_SERVICE_IF_001',whereKey='itemType',whereValue='lineItem')['object']
        C_NOS_EQUIP_GIGA_ROUTER = objectUtil.getSiblingWhere(items,selectKey='ProductCode',selectKeyValue='C_NOS_EQUIP_GIGA_ROUTER',whereKey='itemType',whereValue='lineItem')['object']

        C_NOS_DN = objectUtil.getSiblingWhere(items,selectKey='ProductCode',selectKeyValue='C_NOS_DN',whereKey='itemType',whereValue='lineItem')['object']
        C_NOS_SERVICE_VF_001 = objectUtil.getSiblingWhere(items,selectKey='ProductCode',selectKeyValue='C_NOS_SERVICE_VF_001',whereKey='itemType',whereValue='lineItem')['object']


        C_NOS_OFFER_001['lineItems']['records'].append(C_NOS_EQUIP_TV_006)
        C_NOS_OFFER_001['lineItems']['records'].append(C_NOS_EQUIP_TV_009)
        C_NOS_OFFER_001['lineItems']['records'].append(C_NOS_SERVICE_TV_003)

        C_NOS_OFFER_001['lineItems']['records'].append(C_NOS_SERVICE_IF_001)
        C_NOS_OFFER_001['lineItems']['records'].append(C_NOS_EQUIP_GIGA_ROUTER)

        C_NOS_OFFER_001['lineItems']['records'].append(C_NOS_DN)
        C_NOS_OFFER_001['lineItems']['records'].append(C_NOS_SERVICE_VF_001)
    #    C_NOS_OFFER_001['lineItems']['records'].append(C_NOS_AGG_SERVICE_VF_HFC)

        items = {"records":[C_NOS_OFFER_001]}
        res = CPQ.updateCartItem_api(cartId,items)
        print(restClient.getLastCallTime())
        print(res['messages'][0]['message'])

        incli._main(["-u","NOSDEV","-q", f"select Id,vlocity_cmt__BillingAccountId__c,vlocity_cmt__LineNumber__c,vlocity_cmt__ServiceAccountId__c,vlocity_cmt__Product2Id__r.productCode from orderitem where OrderId='{cartId}' order by vlocity_cmt__LineNumber__c","-fields","all"])  

        print()

    def lineItem_for_update(self,lineItem):
          obj = {}
          obj['vlocity_cmt__LineNumber__c'] = lineItem['vlocity_cmt__LineNumber__c']
          obj['vlocity_cmt__RootItemId__c'] = lineItem['vlocity_cmt__RootItemId__c']
          obj['vlocity_cmt__Action__c'] = lineItem['vlocity_cmt__Action__c']          
          obj['vlocity_cmt__AssetReferenceId__c'] = lineItem['vlocity_cmt__AssetReferenceId__c']
          obj['Id'] = lineItem['Id']

   #       obj['vlocity_cmt__ServiceAccountId__c'] = account2Id
   #       obj['vlocity_cmt__BillingAccountId__c'] = account2Id

          if 'lineItems2' in lineItem:
            obj['lineItems'] = lineItem['lineItems']
            for li in obj['lineItems']['records']:
              li.pop('productChildItemDefinition')
              li.pop('lineItems',None)
              li.pop('childProducts',None)
              li.pop('productGroups',None)

              li.pop('messages',None)
              li.pop('actions',None)
              li.pop('displaySequence',None)
              li.pop('UnitPrice',None)
              li.pop('Product2',None)

              li.pop('Name',None)
              li.pop('Product2Id',None)
              li.pop('IsActive',None)
              li.pop('ProductCode',None)
              li.pop('vlocity_cmt__RecurringPrice__c',None)

              li.pop('Pricebook2Id',None)
              li.pop('PricebookEntry',None)
              li.pop('productId',None)
              li.pop('defaultQuantity',None)
              li.pop('minQuantity',None)

              li.pop('maxQuantity',None)
              li.pop('groupMinQuantity',None)
              li.pop('groupMaxQuantity',None)
              li.pop('sequenceNumber',None)
              li.pop('productChildItemId',None)
              li.pop('productChildItemDefinition',None)

              li.pop('name',None)
              li.pop('isVirtualItem',None)
              li.pop('hasChildren',None)
              li.pop('orderActive',None)
              li.pop('parentRecordTypeDevName',None)
              li.pop('itemType',None)

              li.pop('action',None)
              li.pop('SellingPeriod',None)
              li.pop('ListPrice',None)
              li.pop('OrderId',None)
              li.pop('PricebookEntryId',None)
              li.pop('Quantity',None)

         #     li.pop('vlocity_cmt__AssetReferenceId__c',None)
              li.pop('vlocity_cmt__CatalogItemReferenceDateTime__c',None)
              li.pop('vlocity_cmt__CurrencyPaymentMode__c',None)
              li.pop('vlocity_cmt__EffectiveOneTimeTotal__c',None)
              li.pop('vlocity_cmt__EffectiveRecurringTotal__c',None)
              li.pop('vlocity_cmt__InCartQuantityMap__c',None)
              li.pop('vlocity_cmt__IsChangesAllowed__c',None)

          if 'lineItems' in lineItem:
            obj['lineItems'] = {}
            obj['lineItems']['records'] = []
            for record in lineItem['lineItems']['records']:
              li = self.lineItem_for_update(record)
              obj['lineItems']['records'].append(li)
          return obj
    def test_do_all_puts(self):
        def getObj(lineItem,account2Id):
          obj = {}
          obj['vlocity_cmt__LineNumber__c'] = lineItem['vlocity_cmt__LineNumber__c']
          obj['vlocity_cmt__RootItemId__c'] = lineItem['vlocity_cmt__RootItemId__c']
          obj['vlocity_cmt__Action__c'] = lineItem['vlocity_cmt__Action__c']          
          obj['vlocity_cmt__AssetReferenceId__c'] = lineItem['vlocity_cmt__AssetReferenceId__c']
          obj['Id'] = lineItem['Id']

          obj['vlocity_cmt__ServiceAccountId__c'] = account2Id
          obj['vlocity_cmt__BillingAccountId__c'] = account2Id

          if 'lineItems2' in lineItem:
            obj['lineItems'] = lineItem['lineItems']
            for li in obj['lineItems']['records']:
              li.pop('productChildItemDefinition')
              li.pop('lineItems',None)
              li.pop('childProducts',None)
              li.pop('productGroups',None)

              li.pop('messages',None)
              li.pop('actions',None)
              li.pop('displaySequence',None)
              li.pop('UnitPrice',None)
              li.pop('Product2',None)

              li.pop('Name',None)
              li.pop('Product2Id',None)
              li.pop('IsActive',None)
              li.pop('ProductCode',None)
              li.pop('vlocity_cmt__RecurringPrice__c',None)

              li.pop('Pricebook2Id',None)
              li.pop('PricebookEntry',None)
              li.pop('productId',None)
              li.pop('defaultQuantity',None)
              li.pop('minQuantity',None)

              li.pop('maxQuantity',None)
              li.pop('groupMinQuantity',None)
              li.pop('groupMaxQuantity',None)
              li.pop('sequenceNumber',None)
              li.pop('productChildItemId',None)
              li.pop('productChildItemDefinition',None)

              li.pop('name',None)
              li.pop('isVirtualItem',None)
              li.pop('hasChildren',None)
              li.pop('orderActive',None)
              li.pop('parentRecordTypeDevName',None)
              li.pop('itemType',None)

              li.pop('action',None)
              li.pop('SellingPeriod',None)
              li.pop('ListPrice',None)
              li.pop('OrderId',None)
              li.pop('PricebookEntryId',None)
              li.pop('Quantity',None)

         #     li.pop('vlocity_cmt__AssetReferenceId__c',None)
              li.pop('vlocity_cmt__CatalogItemReferenceDateTime__c',None)
              li.pop('vlocity_cmt__CurrencyPaymentMode__c',None)
              li.pop('vlocity_cmt__EffectiveOneTimeTotal__c',None)
              li.pop('vlocity_cmt__EffectiveRecurringTotal__c',None)
              li.pop('vlocity_cmt__InCartQuantityMap__c',None)
              li.pop('vlocity_cmt__IsChangesAllowed__c',None)

          if 'lineItems' in lineItem:
            obj['lineItems'] = {}
            obj['lineItems']['records'] = []
            for record in lineItem['lineItems']['records']:
              li = getObj(record,account2Id)
              obj['lineItems']['records'].append(li)
          return obj

        restClient.init('NOSDEV')

        accountname='unaiTest_1'

        cartId = self.create_cart_with_promo(accountname=accountname,cartName='testa_11',promoName='NOS4u 100Mb + Móvel',create=True,pricelistF='Name:B2C Price List')

        all = CPQ.getCartItems_api(cartId,fields='vlocity_cmt__ServiceAccountId__c,vlocity_cmt__BillingAccountId__c',price=False,validate=False,includeAttachment=False)
        print(restClient.getLastCallTime())

        account1Id = account.getId(f'name:{accountname}')
        account2Id = account.getId('name:unai2')

     #   strAll = simplejson.dumps(all)
     #   stItems2 = strAll.replace(account1Id,account2Id)
        items = objectUtil.replace_everywhere_in_obj(all,account1Id,account2Id)

        print(cartId)

        codes = ['C_NOS_SERVICE_TV_003','C_NOS_EQUIP_TV_006','C_NOS_EQUIP_TV_009','C_NOS_SERVICE_IF_001','C_NOS_EQUIP_GIGA_ROUTER','C_NOS_DN','C_NOS_SERVICE_VF_001','C_NOS_MSISDN','C_NOS_SERVICE_VM_001','C_NOS_EQUIP_SIM_CARD','C_NOS_OFFER_001','C_NOS_OFFER_006','C_NOS_AGG_SERVICE_TV_UMA','C_NOS_AGG_SERVICE_IF_HFC','C_NOS_AGG_SERVICE_VF_HFC','C_NOS_AGG_SERVICES_MAND_VM']
        codes = ['C_NOS_OFFER_006']
      #  codes = []
        lineItem = objectUtil.getSiblingWhere(items,selectKey='ProductCode',selectKeyValue='C_NOS_OFFER_001',whereKey='itemType',whereValue='lineItem')['object']
   #     for key in lineItem.keys():
   #       print(key)
        C_NOS_OFFER_001 = getObj(lineItem,account2Id)
        C_NOS_OFFER_001['lineItems']['records'] = []
    #    C_NOS_OFFER_001_copy = C_NOS_OFFER_001.copy()
    #    C_NOS_OFFER_001_copy.pop('lineItems')
        if 1==1:
            C_NOS_OFFER_001.pop('vlocity_cmt__LineNumber__c')
            C_NOS_OFFER_001.pop('vlocity_cmt__RootItemId__c')
            C_NOS_OFFER_001.pop('vlocity_cmt__Action__c')
            C_NOS_OFFER_001.pop('Id')
            C_NOS_OFFER_001.pop('vlocity_cmt__ServiceAccountId__c')
            C_NOS_OFFER_001.pop('vlocity_cmt__BillingAccountId__c')
            C_NOS_OFFER_001.pop('vlocity_cmt__AssetReferenceId__c')

    #    C_NOS_OFFER_001['lineItems'] = {}
    #    C_NOS_OFFER_001['lineItems']['records']=[]

    #    C_NOS_OFFER_001['lineItems']['records'].append(C_NOS_OFFER_001_copy)
        for code in codes:
          lineItem = objectUtil.getSiblingWhere(items,selectKey='ProductCode',selectKeyValue=code,whereKey='itemType',whereValue='lineItem')['object']
          li = getObj(lineItem,account2Id)
          C_NOS_OFFER_001['lineItems']['records'].append(li)

        items = {"records":[C_NOS_OFFER_001]}
        res = CPQ.updateCartItem_api(cartId,items)  
        filename = restClient.callSave('records0001',logRequest=True,logReply=False)   

        print(restClient.getLastCallTime())
        print(res['messages'][0]['message'])

        incli._main(["-u","NOSDEV","-q", f"select Id,vlocity_cmt__BillingAccountId__c,vlocity_cmt__LineNumber__c,vlocity_cmt__ServiceAccountId__c,vlocity_cmt__Product2Id__r.productCode from orderitem where OrderId='{cartId}' order by vlocity_cmt__LineNumber__c","-fields","all"])  

    def test_get_order_lines_for_CartId(self):
        cartId = '8013O000003j97nQAA'
        incli._main(["-u","NOSDEV","-q", f"select Id,vlocity_cmt__LineNumber__c,vlocity_cmt__Product2Id__r.productCode,vlocity_cmt__Product2Id__r.name from orderitem where OrderId='{cartId}' order by vlocity_cmt__LineNumber__c","-fields","all"])  

    def test_deleteOrder(self):
      orderId = '8013O000003a9CoQAI'
      restClient.init('NOSDEV')
      try:
        delete = CPQ.deleteCart(orderId)
      except Exception as e:
        utils.printException(e)

    def test_delete_discount(self):

      restClient.init('NOSDEV')
      
    def create_cart_with_promo(self,accountname,cartName,promoName,create=False,pricelistF='Name:B2C Price List'):

        cartId = CPQ.getCartId(f"name:{cartName}")
        if cartId != None:
          if create == False:
            return cartId
          CPQ.deleteCart(cartId)

        accountF = f'Name:{accountname}'
        accountId = account.create_Id(accountname,recordTypeName='Consumer')
        cartId = CPQ.createCart(accountF= accountId, pricelistF=pricelistF,name=cartName,checkExists=True)

     #   promoName = 'NOS4u 100Mb + Móvel'
        promo = CPQ.getCartPromotions(cartId,query=promoName,onlyOne=True)

        res = CPQ.postCartsPromoItems_api(cartId,promo['Id'],price=False,validate=False)

        return cartId

    def test_post_cart_item(self):
        restClient.init('NOSQSM')
        restClient.setLoggingLevel()
        cartName = 'testa_4'

        cartId = CPQ.getCartId(f"name:{cartName}")
        if cartId != None:
          CPQ.deleteCart(cartId)

        accountId = account.createAccount_Id('Name:unaiTest2',recordTypeName='Consumer')
        cartId = CPQ.createCart(accountF= accountId, pricelistF='Name:B2C Price List',name=cartName,checkExists=True)
        print(restClient.getLastCallTime())

        promoName = 'NOS4u 100Mb + Móvel'
        promo = CPQ.getCartPromotions(cartId,query=promoName,onlyOne=True)

        res = CPQ.postCartsPromoItems_api(cartId,promo['Id'])
        print(restClient.getLastCallTime())

        all = CPQ.getCartItems_api(cartId,fields='vlocity_cmt__ServiceAccountId__c,vlocity_cmt__BillingAccountId__c',price=False,validate=False,includeAttachment=False)
        print(restClient.getLastCallTime())

    def test_create_cart_from_asset(self):
        restClient.init('DEVNOSCAT2')

        restClient.setLoggingLevel()
        cartName = 'testa_4'

        cartId = CPQ.createCartFromAsset(assetId='02i3O00000KQ489QAD',accountId='0013O0000186LXLQA2',date='2023-02-22',inputFields={"Name":cartName})

        print()

    def test_update_attribute2(self):
        restClient.init('NOSQSM')
        restClient.setLoggingLevel()
        cartName = 'testa_4'

        promoName = 'NOS4u 100Mb + Móvel'

        cartId = self.create_cart_with_promo('account_2','cart_2',promoName,create=False)
        print(cartId)
        items = CPQ.getCartItems_api(cartId,price=False,validate=False)
        lineItem = objectUtil.getSiblingWhere(items,selectKey='ProductCode',selectKeyValue='C_NOS_SERVICE_TV_003',whereKey='itemType',whereValue='lineItem')['object']
        line2 = self.lineItem_for_update(lineItem)

        def onlyValue(line):
          obj = {}
          obj['value'] = line['value']
          return obj

        line2['vlocity_cmt__LineNumber__c'] = onlyValue(line2['vlocity_cmt__LineNumber__c'])
        line2['vlocity_cmt__RootItemId__c'] = onlyValue(line2['vlocity_cmt__RootItemId__c'])
        line2['vlocity_cmt__AssetReferenceId__c'] = onlyValue(line2['vlocity_cmt__AssetReferenceId__c'])
        line2['Id'] = onlyValue(line2['Id'])
      #  line2['vlocity_cmt__Action__c'] = onlyValue(line2['vlocity_cmt__Action__c'])
        line2['vlocity_cmt__Action__c'].pop('previousValue')
        line2['vlocity_cmt__Action__c'].pop('originalValue')
        line2['vlocity_cmt__Action__c'].pop('messages')
        line2['vlocity_cmt__Action__c'].pop('actions')
        line2['vlocity_cmt__Action__c'].pop('hidden')
      #  line2['vlocity_cmt__Action__c'].pop('editable')  #REQ
      #  line2['vlocity_cmt__Action__c'].pop('dataType')  #REQ
        line2['vlocity_cmt__Action__c'].pop('alternateValues')
      #  line2['vlocity_cmt__Action__c'].pop('label')
       # line2['vlocity_cmt__Action__c'].pop('fieldName')  #REQ


        line2['attributeCategories'] = lineItem['attributeCategories']
        line2['attributeCategories']['records'][0]['productAttributes']['records'][2]['userValues'] = 'yyyy5'
        line2['attributeCategories']['records'].pop()
        line2['attributeCategories']['records'].pop()
        line2['attributeCategories']['records'][0]['productAttributes']['records'].pop(0)
        line2['attributeCategories']['records'][0]['productAttributes']['records'].pop(0)

        line2['attributeCategories']['records'][0]['productAttributes'].pop('totalSize')
        line2['attributeCategories']['records'][0]['productAttributes'].pop('messages')

        line2['attributeCategories']['records'][0].pop('displaySequence')
        line2['attributeCategories']['records'][0].pop('Name')
        line2['attributeCategories']['records'][0].pop('id')
        line2['attributeCategories']['records'][0].pop('messages')

       # line2['attributeCategories']['records'][0].pop('Code__c')  REQ

        line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('messages')
        line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('dataType')
     #   line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('inputType',None)   REQ
        line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('hasRules')
        line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('cloneable')
        line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('multiselect')
        line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('isNotTranslatable')


     #   line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('required')     REQ
        line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('attributeId')

        line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('filterable')
        line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('label')
        line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('displaySequence')

        line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('disabled',None)
        line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('readonly',None)
        line2['attributeCategories'].pop('totalSize')
        line2['attributeCategories'].pop('messages')

        line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('values')
        line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('hidden',None)
        line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('disabled',None)
        line2['attributeCategories']['records'][0]['productAttributes']['records'][0].pop('readonly',None)

        items = {"records":[line2]}
        res = CPQ.updateCartItem_api(cartId,items)  
        filepath = restClient.callSave("update1111",logRequest=True,logReply=False)
        print(res['messages'][0]['message'])
        print(restClient.getLastCallTime())

        itemsAfter = CPQ.getCartItems_api(cartId,price=False,validate=False)

        line3 = objectUtil.getSiblingWhere(itemsAfter,selectKey='ProductCode',selectKeyValue='C_NOS_SERVICE_TV_003',whereKey='itemType',whereValue='lineItem')['object']
        print(line3['attributeCategories']['records'][0]['productAttributes']['records'][2]['userValues'])
        print()

    def test_update_attribute(self):
        restClient.init('NOSQSM')
        restClient.setLoggingLevel()

        cartId = self.create_cart_with_promo('account_2','cart_2','NOS4u 100Mb + Móvel',create=False)

        items = CPQ.getCartItems_api(cartId)

        item = items['records'][1]['lineItems']['records'][0]
        item2 = {}
        item2['attributeCategories'] = item['attributeCategories']
        item2['attributeCategories']['records'].pop()
        item2['attributeCategories']['totalSize'] = 1

        item2['Id'] = item['Id']
        item2['Product2'] = item['Product2']

        item2['PricebookEntry'] = {}
        item2['PricebookEntry']['attributes'] = item['PricebookEntry']['attributes']
        item2['PricebookEntry']['Product2Id'] = item['PricebookEntry']['Product2Id']
        item2['PricebookEntry']['Pricebook2Id'] = item['PricebookEntry']['Pricebook2Id']
        item2['PricebookEntry']['Product2'] = item['PricebookEntry']['Product2']
        item2['PricebookEntry']['Id'] = item['PricebookEntryId']['value']
     #   item2['PricebookEntry']['value'] = item['PricebookEntry']['value']

        item2['PricebookEntry']['Product2'].pop('vlocity_cmt__AttributeDefaultValues__c')
        item2['PricebookEntry']['Product2'].pop('vlocity_cmt__AttributeMetadata__c')

        item2['PricebookEntryId'] = item['PricebookEntryId']

        attributes = CPQ.update_item_attribute_api(cartId,item2)


        print()

    def test_get_cart_items(seld):
        restClient.init('NOSDEV')
        restClient.setLoggingLevel()
        cartId = "8013O000003keqvQAA"

        items = CPQ.getCartItems_api(cartId,price=False,validate=False)

        print()

    def test_submit_order(seld):
        restClient.init('DEVNOSCAT2')
        restClient.setLoggingLevel()
        cartId = "8013O000003izGtQAI"

        items = CPQ.checkOut_api(cartId)

        print()

    def test_post_cart_items(self):
        restClient.init('DEVNOSCAT4')
        restClient.setLoggingLevel()
        pricelistF = "Name:B2C Price List"

        ts = timeStats.TimeStats()

        cartId = self.create_cart_with_promo('account_2','cart_7','NOS4u 100Mb + Móvel',pricelistF=pricelistF,create=True)
        print(cartId)
        items = CPQ.getCartItems_api(cartId,price=False,validate=False)
        items = CPQ.getCartItems_api(cartId,price=False,validate=False)
        items = CPQ.getCartItems_api(cartId,price=False,validate=False)
        items = CPQ.getCartItems_api(cartId,price=False,validate=False)
        items = CPQ.getCartItems_api(cartId,price=False,validate=False)

        req1,res1 = restClient.callSave('getITems123')
        parentRecord = items['records'][0]['productGroups']['records'][0]

        priceBookEntryId = items['records'][0]['productGroups']['records'][0]['childProducts']['records'][0]['PricebookEntry']['Id']
        parentId = items['records'][0]['Id']['value']
        parentHierarchyPath = items['records'][0]['productGroups']['records'][0]['productHierarchyPath']

        parentRecord2 = {}
        parentRecord2['Id'] = parentRecord['Id']['value']
        parentRecord2['productId'] = parentRecord['productId']
        parentRecord2['productHierarchyPath'] = parentRecord['productHierarchyPath']
        parentRecord2['itemType'] = parentRecord['itemType']
        parentRecord2['parentLineItemId'] = parentRecord['parentLineItemId']

        for x in range(0,1):
          ts.new()

          call = CPQ.addItemstoCart_api(cartId,priceBookEntryId,parentRecord=parentRecord,parentId=parentId,parentHierarchyPath=parentHierarchyPath,noResponseNeeded=False,price=False,validate=False)
          ts.time('postCartItem')

          req2,res2 = restClient.callSave("addItemstoCart_api123",logReply=True,logRequest=True)
          
          items2 = CPQ.getCartItems_api(cartId,price=False,validate=False)
          ts.time('getCartItems')

          itemIds = items2['records'][0]['productGroups']['records'][0]['lineItems']['records'][0]['Id']['value']

          call2 = CPQ.deleteCartItems_api(cartId,itemIds=[itemIds],parentRecord=parentRecord2,price=False,validate=False)
          ts.time('deleteCartItem')

        ts.print()
        print()

    def test_getCartItems_x(self):
        restClient.init('NOSDEV')
        q="""SELECT id,
                    pricebook2id,
                    accountid,
                    createddate,
                    vlocity_cmt__defaultcurrencypaymentmode__c,
                    vlocity_cmt__effectiverecurringtotal__c,
                    vlocity_cmt__effectiveonetimetotal__c,
                    vlocity_cmt__numberofcontractedmonths__c,
                    vlocity_cmt__pricelistid__c,
                    recordtypeid,
                    recordtype.developername,
                    status,
                    effectivedate,
                    vlocity_cmt__requestdate__c,
                    vlocity_cmt__orderstatus__c,
                    vlocity_cmt__ischangesallowed__c,
                    vlocity_cmt__ischangesaccepted__c,
                    vlocity_cmt__supplementalaction__c,
                    vlocity_cmt__supersededorderid__c,
                    vlocity_cmt__firstversionorderidentifier__c,
                    vlocity_cmt__requestedstartdate__c,
                    vlocity_cmt__originatingcontractid__c,
                    vlocity_cmt__lastpricedat__c,
                    vlocity_cmt__validationdate__c,
                    vlocity_cmt__ordergroupid__c,
                    vlocity_cmt__ordergroupid__r.vlocity_cmt__groupcartid__c,
                    nos_t_scenariolevel__c,
                    nos_t_coveragetechnology__c,
                    nos_t_businessscenario__c,
                    nos_t_channel__c,
                    nos_t_process__c,
                    nos_b_isbsimulation__c 
                    FROM Order WHERE Id IN ('8013O000003keqvQAA')"""
        res = query.query(q)

        print(restClient.getLastCallTimes())

        q2 =f"""SELECT Id,
                      Quantity,
                      vlocity_cmt__AssetReferenceId__c,
                      vlocity_cmt__ParentItemId__c,
                      vlocity_cmt__RootItemId__c,
                      PriceBookEntry.Product2Id,
                      PriceBookEntry.Name,
                      PriceBookEntry.Product2.vlocity_cmt__GlobalGroupKey__c,
                      vlocity_cmt__LineNumber__c,
                      vlocity_cmt__Product2Id__c,
                      vlocity_cmt__Action__c,
                      PricebookEntryId,
                      vlocity_cmt__ProvisioningStatus__c,
                      vlocity_cmt__ProductHierarchyPath__c,
                      vlocity_cmt__ProductHierarchyGroupKeyPath__c,
                      vlocity_cmt__CatalogItemReferenceDateTime__c,
                      OrderId,
                      vlocity_cmt__SupplementalAction__c,
                      vlocity_cmt__IsChangesAllowed__c 
                      FROM OrderItem WHERE OrderId = '{res['records'][0]['Id']}'  ORDER BY vlocity_cmt__LineNumber__c """
        res2 = query.query(q2)
        print(restClient.getLastCallTimes())

        idl = [r['Id'] for r in res2['records']]
        lineItemIds = query.IN_clause(idl)

        q3 = f"""SELECT vlocity_cmt__OrderAppliedPromotionId__r.vlocity_cmt__PromotionId__c,
                        vlocity_cmt__OrderAppliedPromotionId__r.vlocity_cmt__Action__c,
                        vlocity_cmt__OrderItemId__c 
                        FROM vlocity_cmt__OrderAppliedPromotionItem__c 
                        WHERE vlocity_cmt__OrderItemId__c IN ({lineItemIds})"""
        res3 = query.query(q3)

        print(restClient.getLastCallTimes())
        print()




import unittest
from InCli import InCli
from InCli.SFAPI import account,restClient,digitalCommerce

class Test_DC(unittest.TestCase):
    def test_catalogs(self):
        restClient.init('DEVNOSCAT4')

        catalogs = digitalCommerce.getCatalogs()

        print()

    def test_getOffers(self):
        restClient.init('DEVNOSCAT4')

        catalogs = digitalCommerce.getCatalogs()
        for catalog in catalogs:
            try:
                offers = digitalCommerce.getOfferByCatalogue(catalog['vlocity_cmt__CatalogCode__c'])
                print(f"offers: {len(offers)}")
            except Exception as e:
                print(f"{catalog['vlocity_cmt__CatalogCode__c']}  {e.args[0]['error']}")

        print()

        



from . import restClient,Sobjects,utils,thread
import simplejson

v = 'v51.0'
def query(q):
    action = f"/services/data/{v}/tooling/query/?q={q}"
    call = restClient.callAPI(action)
    checkError()
    for r in call['records']: r.pop('attributes')

    return call

def query_threaded(q,values,search="$$$",raiseEx=True,th=10):
    result = []

    def do_work(value):
        q1 = q.replace(search,value)
        res = query(q1)
        return res

    def on_done(res,result):
        result.append(res['records'][0])

    thread.execute_threaded(values,result,do_work,on_done,threads=th)

    return result

def get(sobject,id):
    action = f"/services/data/{v}/tooling/sobjects/{sobject}/{id}"
    call = restClient.callAPI(action)
    return call

def checkError():
    call = restClient.lastCall()['response']
    if 'serverResponse' in call:
        sr = call.split('serverResponse:')[1]
        srj = simplejson.loads(str(sr))
        utils.raiseException(srj[0]['errorCode'],srj[0]['message'])  

def post(sobject,data):
    action = f"/services/data/{v}/tooling/sobjects/{sobject}"
    call = restClient.callAPI(action,method='post',data=data)
    checkError()
    return call

def delete(sobject,id):
    action = f"/services/data/{v}/tooling/sobjects/{sobject}/{id}"
    call = restClient.callAPI(action,method='delete')
    return call

def patch(sobject,id,data):
    action = f"/services/data/{v}/tooling/sobjects/{sobject}/{id}"
    call = restClient.callAPI(action,method='patch',data=data)
    return call

def idF(object,fieldF,multiple=False):
    chunks = fieldF.split(":")
    if len(chunks)<2:
        utils.raiseException("fieldF error",f"Not a valid fieldF name:value  {fieldF}")
    q = f"select id from {object} where {chunks[0]} = '{chunks[1]}'"
    call = query(q)   
    if len(call['records']) == 0:
        return None
    if multiple:
        return call['records']
    return call['records'][0]['Id']

def describe(sobject):
    action =f"/services/data/{v}/tooling/sobjects/{sobject}/describe/"
    call = restClient.callAPI(action)
    utils.printFormated(call['fields'],"label:name:type")
  #  for field in call['fields']:
  #      print(f"{field['label']}  {field['name']} {field['type']}")
    print()
    

#def get():
#    action = "/services/data/v51.0/tooling/sobjects/TraceFlag/"

def queryTraceFlg(q):
    q = "select id, TracedEntityId,logtype, startdate, expirationdate, debuglevelid, debuglevel.apexcode, debuglevel.visualforce from TraceFlag limit 10"
    call = query(q)
    print()
    



from cmath import log
import logging
import simplejson

ident = -1
treeObjects = []
keyPath = []
def reset():
    global ident,treeObjects
    ident = -1
    treeObjects = []
    keyPath = []

def getChild(json,name):
    for key in json.keys():
        if key == name:
            return json[key]
    return ''

#renaiming and return root Json dict. 
def setKeyValue(json, selectKey, selectKeyValue,setKey, setKeyValue,logValues=False):
    sibling = parse(json, selectKey, selectKeyValue,setKey, setKeyValue,logValues)
    return json

def Id(obj):
    sibbling = getSibling(obj,'Id')
    Id = sibbling['Id']
    if 'value' in Id:
        Id = Id['value']
    return Id

def getSiblingEx(obj, selectKey, selectKeyValue,logValues=False):
    reset()
    retObject = []
    if isinstance(obj,list)==True:
        for element in obj:
            ret = parseEx(element, selectKey=selectKey, selectKeyValue=selectKeyValue,logValues=logValues)
            retObject.append(ret)
        return retObject
    return parseEx(obj, selectKey=selectKey, selectKeyValue=selectKeyValue,logValues=logValues)

def getSibling(obj, selectKey, selectKeyValue='',logValues=False,multiple=False):
    global treeObjects,keyPath
    reset()
    if isinstance(obj,list)==True:
        rets = []

        for element in obj:
            ret = parse(element, selectKey=selectKey, selectKeyValue=selectKeyValue,logValues=logValues,rets=rets)
            if ret != '':
                if multiple is False:
                    return ret
                rets.append(ret)
        if multiple is False:
            return None
        return rets
    p = parse(obj, selectKey=selectKey, selectKeyValue=selectKeyValue,logValues=logValues)
    treeObjects = treeObjects[0:ident+1]
    keyPath = keyPath[0:ident+1]

    return p

def getSiblingWhere(obj, selectKey,selectKeyValue=None, whereKey=None,whereValue=None,logValues=False):

    result = {
        'level':-1,
        'keys':[],
        'objects':[],
        'object':None
    }

    if isinstance(obj,list)==True:
        for element in obj:
            ret = parseWhere(element, selectKey=selectKey,selectKeyValue=selectKeyValue, whereKey=whereKey, whereValue=whereValue,logValues=logValues,result=result)
            if ret['object'] != None:
                return ret
        return None
    ret = parseWhere(obj, selectKey=selectKey, selectKeyValue=selectKeyValue,whereKey=whereKey, whereValue=whereValue,logValues=logValues,result=result)
    if ret['level']<0:
        ret['level'] = 0
    ret['keys'] = ret['keys'][0:ret['level']]
    ret['objects'] = ret['objects'][0:ret['level']]

    return ret
#------------------------
def check(json,text):
    found = False
    data = simplejson.dumps(json, indent=2)
    if text in data:
        print(text)

def getValue(obj,name):
    value = obj[name]
    if 'value' in obj[name]:
        value = obj[name]['value']  
    return value

#set selecte key where keyvalue is selectKeyValue and return jsonDic 
# if set Key provided, will replace the setKeyValue
def parseWhere(json,result,selectKey='', selectKeyValue=None,whereKey=None, whereValue=None,logValues=False):
    def setret(result,json):
        result['object'] = json
        return result      

    result['level']=result['level']+1
    result['objects'].insert(result['level'], json)

    if isinstance(json,dict)==False and isinstance(json,list)==False:
        return setret(result,None)

    for key in json.keys():
        result['keys'].insert(result['level'],key)

        if key == selectKey:
            if selectKeyValue == None:
                if whereKey == None:
                    return setret(result,json[key])

            else:
                value = getValue(json,selectKey) 
                if value == selectKeyValue:
                    if whereKey == None:
                        return setret(result,json)
                    elif whereKey in json:
                        if json[whereKey] == whereValue:
                            return setret(result,json)

        if isinstance(json[key],dict):
            ret = parseWhere(json[key],result,selectKey, selectKeyValue,whereKey,whereValue,logValues)

            if ret['object'] != None:
                return ret

        if isinstance(json[key],list):
            for l in json[key]:
                ret = parseWhere(l,result,selectKey,selectKeyValue,whereKey,whereValue,logValues)
  
                if ret['object'] != None:
                    return ret
    result['level'] = result['level']-1

    result['object'] = None
    return result

def parse(json, selectKey='', selectKeyValue='',setKey='', setKeyValue='',logValues=False,rets=None):
    global ident,treeObjects
    ident=ident+1
    treeObjects.insert(ident, json)
#    print(ident)


    if isinstance(json,dict)==False and isinstance(json,list)==False:
        return ''  #continue

    for key in json.keys():
        keyPath.insert(ident,key)
        if key == 'PricebookEntry':
            continue

        if key == 'ProductCode' and logValues == True:
            printIdent(f'{key}  {json[key]}')

        if key == selectKey:
            if selectKeyValue == '':
                return json
            value = getValue(json,selectKey) 

         #   printIdent(f'{key}  {value}')

            if value == selectKeyValue:
                if setKey!='':
                    if setKeyValue == None:
                        return json[setKey]
                    if json[setKey] != None and type(json[setKey]) is dict and 'value' in json[setKey]:
                        json[setKey]['value'] = setKeyValue
                    else:
                        json[setKey] = setKeyValue
           #     ident = ident-1

                return json 

        if isinstance(json[key],dict):
            ret = parse(json[key],selectKey, selectKeyValue,setKey,setKeyValue)
          #  ident = ident-1

            if ret != '':
                return ret if rets is None else rets.append(json)

        if isinstance(json[key],list):
            for l in json[key]:
                ret = parse(l,selectKey,selectKeyValue,setKey,setKeyValue)
            #    ident = ident-1
  
                if ret != '':
                    return ret if rets is None else rets.append(json)
    ident = ident-1

    return ''

#set selecte key where keyvalue is selectKeyValue and return jsonDic 
# if set Key provided, will replace the setKeyValue
def parseEx(json, selectKey='', selectKeyValue='',setKey='', setKeyValue='',logValues=False,retObjects=[]):
    global ident
    ident=ident+1
  
    #check(json,"AttributeCategory")
    #check(json,"attributeCategories")

    if isinstance(json,dict)==False and isinstance(json,list)==False:
        return ''  #continue

    for key in json.keys():
        if key == 'PricebookEntry':
            continue

        if key == 'ProductCode' and logValues == True:
            printIdent(f'{key}  {json[key]}')

        if key == selectKey:
            value = json[selectKey]
            if 'value' in json[selectKey]:
                value = json[selectKey]['value']
            if value == selectKeyValue:
                if setKey!='':
                    if setKeyValue == None:
                        return json[setKey]
                    if json[setKey] != None and type(json[setKey]) is dict and 'value' in json[setKey]:
                        json[setKey]['value'] = setKeyValue
                    else:
                        json[setKey] = setKeyValue
                #return json     
                retObjects.append(json[selectKey])

        if isinstance(json[key],dict):
            ret = parse(json[key],selectKey, selectKeyValue,setKey,setKeyValue,retObjects)
            if ret != '':
                retObjects.append(ret)
                #return ret
            ident = ident-1

        if isinstance(json[key],list):
            for l in json[key]:
                ret = parse(l,selectKey,selectKeyValue,setKey,setKeyValue,retObjects)
                if ret != '':
                    retObjects.append(json)
                    #return ret
                ident = ident-1
    return retObjects

def addNode(path,node,level):
    if len(path) > level:
        path[level] = node
    else:
        path.append(node) 

def getField(obj,path,separator=':'):
    """
    Get field in object for a path. 
    - path: the path
    - separator: for the path. a:b:c by default. 
    """
    paths = path.split(separator)
    _obj = obj
    for p in paths:
        if p in _obj:
            _obj = _obj[p]
        else:
            return None
    return _obj

def getpaths_toField(json,fieldValue,level=0,paths=[],path=[]):
    try:
        if type(json) is dict:
            for key in json.keys():
                addNode(path,key,level)

                if isinstance(json[key],dict):
                    getpaths_toField(json[key],fieldValue,level+1,paths,path)
                if isinstance(json[key],list):
                    for element in json[key]:
                        getpaths_toField(element,fieldValue,level+1,paths,path)
                if (json[key]==fieldValue):
                    addNode(path,key,level+1)
                    path =  path[0:level+1]
                    str = ".".join(path)
                    paths.append(str)
    except Exception as e:
        logging.error(f'{e}')
    return paths

def printIdent(string):
    global ident
    str = ''
    for x in range(ident):
        str = str + ' '
    print(str + string)


def getNode(obj,path):
    if (path == ""):
        return obj
    
    nodeNames = path.split(".")

    for nodeName in nodeNames:
        if '*' in nodeName:
            print()
        if nodeName.isnumeric():
            index = int(nodeName)
            if index>len(obj):
                return "Error"
            obj = obj[index]

        else:
            if nodeName in obj:
                obj = obj[nodeName]
            else:
                return "Error: " + nodeName + " not present in Object"

    return obj


def replace_everywhere_in_obj(obj,find,replace):
    strAll = simplejson.dumps(obj)
    stItems2 = strAll.replace(find,replace)
    return simplejson.loads(stItems2)





from . import restClient,utils,query,thread
import simplejson
#------------------------------------------------------------------------------------------------------------------------------------------------
def dr(bundleName,inputData):
   """
   - bundleName: Name of the Data Raptor Bundle
   - inputData: input Data to the DR
   """
   action = f'/services/apexrest/{restClient.getNamespace()}/v2/DataRaptor'

   data = {
       "bundleName" : bundleName, 
       "objectList": inputData
   }
   call = restClient.callAPI(action, method="post", data=data)
   return call
#------------------------------------------------------------------------------------------------------------------------------------------------
def ipResponse(name,input,options=None,field=None):
   """
   Calls the IP and returns the response object
   - name: type_subtype
   - input: input data json.
   - options: the options json. 
   - field: if specified will return the value of the 'field' in response.  
   """
   res = ip(name,input,options)
   IPResult = res['IPResult']
   if 'response' in IPResult:
      IPResult = IPResult['response']

   if field != None and field in IPResult:
      return IPResult[field]
   return IPResult
#------------------------------------------------------------------------------------------------------------------------------------------------
def ip(name,input,options=None):
   """
   Calls the IP 
   - name: type_subtype
   - input: input data json.
   - options: the options json. 
   """
   action = f'/services/apexrest/vlocity_cmt/v1/GenericInvoke/vlocity_cmt.IntegrationProcedureService/{name}'
   data = {
      "sMethodName":name,
      "sClassName":"vlocity_cmt.IntegrationProcedureService",
      }

   if type(input) is dict:
      input = simplejson.dumps(input)

   data['input'] = input
   if options == None:
      options = "{}"
   else:
      if type(options) is dict:
         options = simplejson.dumps(options)      
   data['options'] =  options


   call = restClient.callAPI(action=action,data=data,method='post')
   lc = restClient.callSave('data1234',logRequest=True,logReply=False)
   utils.printJSON(call)
   #print(call)
   return call
#------------------------------------------------------------------------------------------------------------------------------------------------
def ip_response(call,field=None):
   """
   Get the response from the IP call response. 
   """
   IPResult = call['IPResult']
   if 'response' in IPResult:
      IPResult = IPResult['response']

   if field != None and field in IPResult:
      return IPResult[field]
   return IPResult
#------------------------------------------------------------------------------------------------------------------------------------------------
def get_IP_definitions():
   """
   Get the IP definitions. 
   Returns the procedure name, the IP name and the steps as strings. 
   """
   q = f"""select 
               Id,
               vlocity_cmt__Content__c,
               vlocity_cmt__OmniScriptId__r.name,
               vlocity_cmt__OmniScriptId__r.vlocity_cmt__ProcedureKey__c 
               from vlocity_cmt__OmniScriptDefinition__c 
               where vlocity_cmt__OmniScriptId__c in (select Id from vlocity_cmt__OmniScript__c where vlocity_cmt__OmniProcessType__c = 'Integration Procedure' and vlocity_cmt__IsActive__c = TRUE) """

   res = query.query(q)

   ip_definitions = []

   for record in res['records']:
      ip_definition = {
            'vlocity_cmt__ProcedureKey__c': record['vlocity_cmt__OmniScriptId__r']['vlocity_cmt__ProcedureKey__c'],
            'Name':                         record['vlocity_cmt__OmniScriptId__r']['Name'],
            'steps':                        []
      }
      ip_definitions.append(ip_definition)
      content = simplejson.loads(record['vlocity_cmt__Content__c'])
      for child in content['children']:
            ip_definition['steps'].append(child['name'])

   return ip_definitions

def call_chainable_ip(name,input,options=None):
   action = f'/services/apexrest/callipasynch/v1/{name}'
   action = f'/services/apexrest/callip/v1/{name}'
   action = f'/services/apexrest/callip/v2/{name}'
   action = f'/services/apexrest/callip/v3/{name}'

   data = {
      "sMethodName":name,
      "sClassName":"vlocity_cmt.IntegrationProcedureService",
      }

   data['input'] = input
   if options == None:
      options = {}
   data['options'] =  options


   call = restClient.callAPI(action=action,data=data,method='post')
   return call
#------------------------------------------------------------------------------------------------------------------------------------------------
def remoteClass(className,method,input,options=None):
   """
   Calls a remote APEX class inside Salesforce 
   - className: the callable APEX class name
   - method: the method to invoke.
   - input: the options json. 
   - options: the options json 
   """
   action = f'/services/apexrest/vlocity_cmt/v1/GenericInvoke/{className}/{method}'

   if type(input) is dict:
      input = simplejson.dumps(input)

   if options == None:
      options = {
         "postTransformBundle":"",
         "preTransformBundle":"",
         "useQueueableApexRemoting":False,
         "ignoreCache":False,
         "vlcClass":"B2BCmexAppHandler",
         "useContinuation":False
         }         
      
   data = {
      "sMethodName":method,
      "sClassName":className,
      "input":input,
      "options":simplejson.dumps(options)
      }

   call = restClient.callAPI(action=action,data=data,method='post')
   return call
#------------------------------------------------------------------------------------------------------------------------------------------------



import unittest
from InCli.SFAPI import restClient,CPQ,account,Sobjects,utils,query,jsonFile,debugLogs
import traceback


class Test_Parse(unittest.TestCase):

    def isRep(self,repeatingSequences,x,parsedLine):
      if parsedLine['type'] != 'METHOD':
        return False,None

      for reps in repeatingSequences:
        delta = reps[1]-reps[0]
        if x>= reps[0] and x <= reps[0] + len(reps) * delta - 1:
          if x>= reps[0] and x <= reps[0] + delta-1:
            parsedLine['output'] = f"{parsedLine['output']}  *** {len(reps)}"
            return True,parsedLine
          else:
            return True,None

      return False,None


    def test_findRepeating(self):

      #  objs = jsonFile.read('/Users/uormaechea/Documents/Dev/python/VirtualEnvs/prj1/.incli/logs/07L3O00000Dg1OSUAZ')

      try:
        logId = '07L3O00000Dg1OSUAZ'
        parseContext = {
          'filepath':f'/Users/uormaechea/Documents/Dev/python/VirtualEnvs/prj1/.incli/logs/{logId}.log'
        }
        parsed = debugLogs.do_parse_from_file(parseContext) 

        repeatingSequences = utils.get_all_loops(parsed,"output")

        parsedX = []
        for x,parsedLine in enumerate(parsed):
          isrepe, obj = self.isRep(repeatingSequences,x,parsedLine)
          if isrepe == True and obj == None:
            continue
          parsedX.append(parsedLine)
        debugLogs.printDebugList(parsedX,logId)    
      

      except Exception as e:
        print(e)
        print(traceback.format_exc())

    def test_do_parse_storage(self):
      try:
        restClient.init('ConnectionLess')
        debugLogs.do_parse_storage()
      except Exception as e:
        print(e)
        traceback.format_exc()

      print()
        






import logging
from . import utils
from . import restClient as client
from . import query
import re,simplejson

apiVersion = 'v51.0'

#X can be an Id or the field value. If it is an Id, it is returned. If not, X it the value of the 'field'. 

def getFieldList(objs,name):
    """
    From and array of objects, return an array of field by name. Return a column. 
    """
    return [obj[name] for obj in objs]

def Id(obj):
    if type(obj) is str:
        return obj
    if type(obj) is dict:
        if 'Id' in obj:
            id = obj['Id']
        if 'id' in obj:
            id = obj['id']
        if 'value' in id:
            id = id['value']
        return id

def getF(objName,extendedF):
    if extendedF == None:
        return None
    if type(extendedF) is dict:
        extendedF = extendedF['Id']
    ef = utils.extendedField(extendedF)
    return query.query(f" select fields(all) from {objName} where {ef['field']} = '{ef['value']}' limit 200")

def IdF(objName,extendedF,init=None):
    """Returns the Id for the especified extendedF in the form "fieldName:fieldValue" after querying SF.
    - In some cases the extendedF contains the Id itself and no query is required.
    - if extendedF is a string, and looks like an Id returns the Id, else exception
    - if extendedF is a dictionary and contains a field Id and looks like an Id, return the Id, else an exception
    - if extendedF is a fieldName:fieldValue, returns the query on object where fieldName=fieldValue
    -
    - objName (String): any SF Object name Order or Account
    - extendedF: a String, and dictionary or a "fieldName:fieldValue"
    - init: To validad the Id (extra checks), the first 3 numbers of the Id. In case of account is 001... in other cases is not as clear cut. A valid Id is checked based on a regex: length of the field and alphanumeric. Thus, It is not 100% error free. Init increases the accuracy. 
    """
    if extendedF == None:
        return None
    if type(extendedF) is dict:
        extendedF = extendedF['Id']

    ef = utils.extendedField(extendedF)
    if ef['field'] == 'Id':
        if checkId(ef['value']):
            if init == None or ef['value'].startswith(init):
                return ef['value']
        utils.raiseException("No_Id",f"{extendedF} is not a valid Id")
    return query.queryField(f" select Id from {objName} where {ef['field']} = '{ef['value']}' ")

def checkId(id,init=None):
    #return re.search(r"[a-zA-Z0-9]{15}|[a-zA-Z0-9]{18}", id)
    if init != None:
        if id.startwith(init) == False:
            return False
    return re.search(r"\b[a-z0-9]\w{4}0\w{12}|[a-z0-9]\w{4}0\w{9}\b", id)

def isId(sobjectType,X):
    if X == None:
        return False
    if type(X) is dict:
        return isId(utils.Id(X))
    if sobjectType.lower() == str(getSObjectType(X)).lower():
        return True   
    return False

#---------------------------------
def getSObjectType(id):
    if checkId(id) == None:
        return None
    if id.startswith("a3O"):
        return "vlocity_cmt__PriceList__c"
    if id.startswith("01s"):
        return "Pricebook2"
    if id.startswith("01t"):
        return "product2"
    if id.startswith("a3P"):
        return "vlocity_cmt__PricingElement__c"
    if id.startswith("a4d"):
        return "vlocity_cmt__PricingElement__c"
    if id.startswith("a36"):
        return "vlocity_cmt__Promotion__c"
    if id.startswith("a3S"):
        return "vlocity_cmt__PromotionItem__c"
    if id.startswith("a3i"):
        return "vlocity_cmt__PriceListEntry__c"
    if id.startswith("a3R"):
        return "vlocity_cmt__PricingVariable__c"
    if id.startswith("a4h"):
        return "vlocity_cmt__PricingVariable__c"
    if id.startswith("a1b"):
        return "vlocity_cmt__ProductChildItem__c"
    if id.startswith("a2S"):
        return "vlocity_cmt__ObjectClass__c"
    if id.startswith("a4M"):
        return "vlocity_cmt__PricingPlan__c"
    if id.startswith("a3T"):
        return "vlocity_cmt__TimePlan__c"
    if id.startswith("a3l"):
        return "vlocity_cmt__TimePolicy__c"
    if id.startswith("001"):
        return "Account"     
    if id.startswith("801"):
        return "Order"    
    if id.startswith("802"):
        return "OrderItem"    
    if id.startswith("02i"):
        return "Asset"    
    if id.startswith("01s"):
        return "Pricebook2" 
    if id.startswith("a0I"):
        return "vlocity_cmt__AttributeCategory__c" 
    if id.startswith("a1B"):
        return "vlocity_cmt__AttributeCategory__c" 
    if id.startswith("a1W"):
        return "vlocity_cmt__Picklist__c" 
    if id.startswith("a4W"):
        return "vlocity_cmt__Picklist__c" 
    if id.startswith("a2m"):
        return "vlocity_cmt__PriceList__c" 
    if id.startswith("a4a"):
        return "vlocity_cmt__PriceList__c" 
    if id.startswith("a2j"):
        return "vlocity_cmt__PicklistValue__c"      
    if id.startswith("a3I"):
        return "vlocity_cmt__ContextDimension__c" 
    if id.startswith("a3b"):
        return "vlocity_cmt__ContextMapping__c" 
    if id.startswith("a3r"):
        return "vlocity_cmt__ContextScope__c" 
    if id.startswith("a3q"):
        return "vlocity_cmt__ContextAction__c" 
    if id.startswith("a0w"):
        return "vlocity_cmt__EntityFilter__c"   
    if id.startswith("a1o"):
        return "vlocity_cmt__Rule__c"        
    if id.startswith("a1l"):
        return "vlocity_cmt__RuleFilter__c"        
    if id.startswith("a41"):
        return "vlocity_cmt__RuleAssignment__c"        
    if id.startswith("a4x"):
        return "vlocity_cmt__OfferMigrationPlan__c"         
    if id.startswith("a4"):
        return "vlocity_cmt__OfferMigrationComponentMapping__c"       
    if id.startswith("005"):
        return "User"   
    if id.startswith("a2H"):
        return "vlocity_cmt__Catalog__c"   
    if id.startswith("a1j"):
        return "vlocity_cmt__Catalog__c"  
    if id.startswith("a04"):
        return "vlocity_cmt__CatalogProductRelationship__c"   
    if id.startswith("a5b"):
        return "vlocity_cmt__ServicePoint__c"        

    print(f" getObjectType--> {id} no related to object")
    return None
#---------------------------------

def get_with_only_id(Id):
    objs= get_sobject_names_and_prefixes()
    prefix = Id[0:3]
    obj = [o for o in objs if o['keyPrefix'] == prefix][0]

    res = get(id=Id,sobjectName= obj['name'])
    return res

def get(id,sobjectName=None):
    if sobjectName == None:
        sobjectName = getSObjectType(id)
    # return select_wherexx_field_value_n(objectName,'Id',id)
    return query.query(f" select fields(all) from {sobjectName} where Id='{id}' limit 200")

#---------------------------------
def create(sobjectname,object):
    call =  client.callAPI(f'/services/data/{apiVersion}/sobjects/{sobjectname}/', method="post", data=object) 
    if client.thread_lastCall()['status_code'] == 201:
        logging.info(f"Object {sobjectname} created sucesfully. Id:{call['id']}")
    else:
        logging.warning(f"Object {sobjectname} creation returned status code {client.thread_lastCall()['status_code']}")
        logging.warning(f"Response {call}")

        raise ValueError(simplejson.dumps(call, indent=4))


    return call

def checkError():
    lc = client.thread_lastCall()
    if lc['status_code'] >= 400:
        logging.error(f"Error in call.  statusCode->{client.thread_lastCall()['status_code']}")
        message = {
            'error':lc['error'],
            'errorCode':lc['errorCode']
        }
        if 'errorOther' in lc:
            message['errorOther'] = lc['errorOther']
        msg = simplejson.dumps(message, indent=4)
        logging.error(msg)
        raise ValueError(msg)

def update(id,data,sobjectname=None,getObject=True):
    if sobjectname == None:
        sobjectname = getSObjectType(id)
    
    call = client.callAPI(f'/services/data/v51.0/sobjects/{sobjectname}/{id}/',method='patch',data=data)
    checkError()
    if getObject == True:
        call = get(id)
    return call

def checkError():
    lc = client.lastCall()

    if 'error' in lc and lc['error'] != None and len(lc['error'])>10:
        if 'response' in lc:
            if 'serverResponse:' in lc['response']:
                s = '{"a":' + lc['response'].split('serverResponse:')[-1] + "}"
                obj = simplejson.loads(s)
                utils.raiseException(obj['a'][0]['errorCode'],obj['a'][0]['message'])
        utils.raiseException(lc['errorCode'],lc['error'])

def deleteMultiple(sobjectname,id_list): 
    def divide_chunks(l, n):
        for i in range(0, len(l), n):
            yield l[i:i + n]
    if len(id_list) == 0: return None
    if len(id_list)>200:
        id200s= list(divide_chunks(id_list,200))
        for id200 in id200s:
            res = deleteMultiple(sobjectname,id200)
        return res

    idList = ",".join(id_list)
    logging.debug(f"deleting {sobjectname} with Id {idList}")
    call = client.callAPI(f'/services/data/{apiVersion}/composite/sobjects?ids={idList}', method="delete")
    try:
        checkError()
    except Exception as e:
        if e.args[0]['errorCode'] == 'ENTITY_IS_DELETED':
            client.glog().info(f"{sobjectname} with Id {id} is already deleted.")
            return None
        else:
            raise e
    return call 

def delete(sobjectname,id):
    if id==None:
        return None
    if type(id) is list:
        return deleteMultiple(sobjectname,id)
    id = Id(id)
   
    logging.debug(f"deleting {sobjectname} with Id {id}")
    action = f'/services/data/{apiVersion}/sobjects/{sobjectname}/{id}'
    print(action)
    call =  client.callAPI(action, method="delete")
    try:
        checkError()
    except Exception as e:
        if e.args[0]['errorCode'] == 'ENTITY_IS_DELETED':
            client.glog().info(f"{sobjectname} with Id {id} is already deleted.")
            return None
        else:
            raise e
    return call

def listObjects():
    return client.callAPI(f'/services/data/{apiVersion}/sobjects/')['sobjects']

def describe(sobjectName):
    return client.callAPI(f'/services/data/{apiVersion}/sobjects/{sobjectName}/describe')

def listVersions():
    return client.callAPI(f'/services/data')

def recordCount(objectsCommaSeparated):
    return client.callAPI(f'/services/data/{apiVersion}/limits/recordCount?sObjects={objectsCommaSeparated}')

def getOwner(Id):
    recordType = getSObjectType(Id)
    if recordType == None:
        return

    select = f"select OwnerId from {recordType} where Id = '{Id}' limit 50"    
    call = query.query(select)
    OwnerId = call['records'][0]['OwnerId']

    select = f"select FIELDS(ALL) from user where Id = '{OwnerId}' limit 50"    
    user = query.query(select)['records'][0]

    print(user['Name'])

    return call

def getNumOwnedRecords(sobject,ownerId):
    try:
        select = f"select COUNT() from {sobject} where OwnerId = '{ownerId}' "    
        call = query.query(select)

        totalSize = call['totalSize']

        return totalSize

    except Exception as e:
        if 'INVALID_FIELD' in f"{e}":
            return -1
    
    return -2

def getNumOwnedRecords_4_ObjectList(sobjectNameList,ownerId):
    result=[]
    for obj in sobjectNameList:
        res = {
            "object":obj,
            "ownedRecords":getNumOwnedRecords(obj,ownerId)
        }
        result.append(res)

    return result

def selectOwnedRecords(sobjName,ownerId):
    select = f"select Fields(ALL) from {sobjName} where ownerId='{ownerId}'limit 50"
    call = query.query(select)

    return call

sobjs=[]
def get_sobject_names_and_prefixes():
    global sobjs
    if len(sobjs) > 0:
        return sobjs
    action = '/services/data/v51.0/sobjects'
    res = client.callAPI(action)

    sobjs=[]
    for obj in res['sobjects']:
        if obj['keyPrefix'] != None:            
            sobj= {
                'name':obj['name'],
                'keyPrefix':obj['keyPrefix']
            }
            sobjs.append(sobj)
    return sobjs


import logging
import threading,time
from queue import Queue
import math
import sys


_counter = 0
#---------------------------------
def getCounter():
  global _counter
  return _counter
#---------------------------------
def count():
  return _counter
  
def printCounter():
  global _counter

  print(_counter)
#---------------------------------
def _testMultiple(func,iterations,processNumber,**funcArg):
  global _counter
  for a in range(iterations):
    func(processNumber=processNumber+a)
    with threadLock:
      _counter += 1
#---------------------------------
threadLock = threading.Lock()
def run(func,iterations,processStartNumber=0,threads=50,onFinished=None):
  global _counter
  _counter = 0

  try:
    for _itera in range(threads):
      processNumber = processStartNumber+ (_itera*threads)
      threading.Thread(target=_testMultiple,args=(func,iterations,processNumber), daemon=True).start()

  except:
    print ("Error: unable to start thread")
  while _counter < iterations * threads :
    logging.info(f'Iteration {_counter}  from {iterations * threads}')
    time.sleep(30)

  if onFinished!=None:
    onFinished()
    
  print("Done")

#---------------------------------
_processed = 0
_totalToProcess = 0

def processList(func,theList,threads):
  global _processed,_totalToProcess

  _processed = 0
  _totalToProcess = len(theList)
  itemsPerTh = math.ceil(_totalToProcess/threads)


  total = 0
  while total < _totalToProcess:
    end = total+itemsPerTh
    if end>_totalToProcess:
      end = _totalToProcess
    data = {
      "list" : theList[total:end]
    }

    threading.Thread(target=_processList,args=(func,theList[total:end],)).start()
    total = total + itemsPerTh

 # print()

  while (_totalToProcess>_processed):
   # print(_processed)
 #   sys.stdout.write("\r%d%%" % int(100*_processed/_totalToProcess))
 #   sys.stdout.flush()
    time.sleep(1)

def progress():
    global _processed,_totalToProcess

    sys.stdout.write("\r%d%%" % int(100*_processed/_totalToProcess))
    sys.stdout.flush() 
    if (_processed==_totalToProcess):
      print()


def _processList(func,ls):
  global _processed
  for l in ls:
    func(l)
   # print('one done')
    with threadLock:
      _processed += 1
      progress()
 # print('X done')
###############
###############
###############

executed_records = 0
total_records = 0
counter_look = threading.Lock()

def increase_counter():
  global executed_records,counter_look,total_records
  with counter_look:
    executed_records = executed_records + 1
    sys.stdout.write("\r%d%%" % int(100*executed_records/total_records))
    sys.stdout.flush() 

def on_work_done(queue,on_done,output_records):
  while True:
    record = queue.get()
    #print(queue.qsize())
    if type(record) is str and record == 'STOP': break
    on_done(record,output_records)
    increase_counter()

def working_thread(queue_in,queue_out,do_work):
  while True:
    record = queue_in.get()
    if type(record) is str and record == 'STOP': break
    ret = do_work(record)
    queue_out.put(ret)


def execute_threaded(input_records,output_records,do_work,on_done,threads=1):
  global executed_records,total_records

  q_in = Queue(maxsize=0)
  q_done = Queue(maxsize=0)

  executed_records = 0
  total_records = len(input_records)

  print(f"Executing threaded records->{total_records}")

  thread_work_done = threading.Thread(target=on_work_done,args=(q_done,on_done,output_records), daemon=True)
  thread_work_done.start()

  all_threads = []
  for t in range(0,threads):
    thread = threading.Thread(target=working_thread,args=(q_in,q_done,do_work), daemon=True)
    thread.start()
    all_threads.append(thread)

  for record in input_records:
    q_in.put(record)

  for x in range(0,threads):
    q_in.put('STOP')

  for thread in all_threads:
    thread.join()

  q_done.put('STOP')  
  thread_work_done.join()
  print()
  print(f"Execution finished threaded records->{len(input_records)}")



from . import Sobjects,utils,restClient,query
import logging

#if the account does not exist, accountX will be used as the account name. 
def getId(accountF):
    """Returns the account Id provided the accountF -> fieldname:fielvalue."""
    Id = Sobjects.IdF('Account',accountF,init='001')
    return Id

def create_Id(accountName,fields=None,recordTypeName=None,checkExists=True):
    """Creates an account
    - accountName - name for the account
    - fields - fields to be set in the account
    - recordType - String with record Type name
    - checkExists - will query if there is an account with the provided accountF - fieldname - The many accounts with the same name, one is returned.
    - returns the accountId
    """
    if checkExists == True:
        accountId = getId(f"Name:{accountName}")

        if accountId != None:
            restClient.glog().info(f'Returning existing Account <{accountId}> ')
            return accountId
            
    acc={
        'Name': accountName,
        'vlocity_cmt__Active__c':'Yes',
        'vlocity_cmt__Status__c':'Active'
    }
    if fields!= None:
        acc.update(fields)

    if recordTypeName!=None:
        acc['RecordTypeId'] = query.queryField(f"select Id from RecordType where SobjectType='Account' and Name = '{recordTypeName}'")

    call = Sobjects.create('Account',acc)

    accountId = call['id']
    restClient.glog().info(f"Account {accountName} Created Account <{accountId}> ")

    return accountId

def create(accountName,fields=None,recordTypeName=None,checkExists=True):
    """Creates an account
    - accountName - name for the account
    - fields - fields to be set in the account
    - recordType - String with record Type name
    - checkExists - will query if there is an account with the provided accountF - fieldname - The many accounts with the same name, one is returned.
    - returns the new account object
    """
    accountId = create_Id(accountName,fields=fields,recordTypeName=recordTypeName,checkExists=checkExists)
    account = Sobjects.get(accountId)

    return account['records'][0]

def get(accountF,multiple=False):
    """
    - multiple - returns an array of accounts (limit 200), else return one account
    - in both cases if no account, returns None
    """
    ef = utils.extendedField(accountF)
    call =  query.query(f" select fields(all) from Account where {ef['field']}='{ef['value']}' limit 200")
    if len(call['records']) == 0:
        return None
    if multiple:
        return call['records']
    return call['records'][0]

def delete(accountF):
    """deletes one account. if multiple match only one is deleted"""
    id = getId(accountF)
    call = Sobjects.delete('Account',id)
    restClient.glog().info(f"Account {accountF} deleted.")
    return call

def getOrdersId(accountF):
    """Returns the orders for the account as a List of Ids
    - if no orders for the account, and empty list is returned."""
    accountId = getId(accountF)
    ids = query.queryFieldList(f" select Id from Order where  AccountId = '{accountId}' ")
    return ids

def deleteOrders(accountF):
    """deletes all the orders for the account. if multiple accounts only the orders for one are deleted"""
    ids = getOrdersId(accountF)
    if ids == [] or ids == None:
        restClient.glog().info(f"Account {accountF} has no orders to delete.")
        return None
    call = Sobjects.delete('Order',ids)
    restClient.glog().info(f"Account {accountF} Orders {ids} deleted.")

    return call



import unittest,simplejson
from InCli import InCli
from InCli.SFAPI import VlocityErrorLog,restClient

class Test_VlocityErrorEntry(unittest.TestCase):

    def test_get_errors(self):
        restClient.init('NOSPRD')

        VlocityErrorLog.get_errors()

        print()

    def test_print_errors(self):
        restClient.init('NOSDEV')

        VlocityErrorLog.print_error_records()

        print()
    def test_get_error_ip(self):
        restClient.init('NOSDEV')

        Id = 'a6K3O000000FHuqUAG'

        q = f"select fields(all) from vlocity_cmt__VlocityErrorLogEntry__c where Id='{Id}' limit 10"

        res = query.query(q)


        out=[]
        for record in res['records']:
            data = record['vlocity_cmt__InputData__c']
            datas = simplejson.loads(data)

            theFile =jsonFile.write('TheIPError_1',datas)
            
            out.append(datas)
            print(datas['ErrorMessage__c'])


      #  file_csv.write('VlocityErrorLogEntry_1',out)

        print()

        



import unittest
from InCli import InCli
from InCli.SFAPI import file,restClient,DR_IP,CPQ,digitalCommerce

class Test_CartItems(unittest.TestCase):
    def test_get_cart_items(self):

        restClient.init('NOSDEV')
        inp = {
            'orderId':"8013O000003keqvQAA"
        }
        res=DR_IP.remoteClass('testGetCartItems','method',inp,{})
        print(restClient.getLastCallTime())
        print(res)

        print()



#Id, ListPrice, OrderId, PricebookEntryId, Product2Id, Quantity, UnitPrice, pricebookentry.pricebook2id, pricebookentry.product2.id, pricebookentry.product2.name, pricebookentry.product2.productcode, pricebookentry.product2.vlocity_cmt__attributedefaultvalues__c, pricebookentry.product2.vlocity_cmt__attributemetadata__c, pricebookentry.product2.vlocity_cmt__globalgroupkey__c, pricebookentry.product2.vlocity_cmt__isconfigurable__c, pricebookentry.product2.vlocity_cmt__type__c, pricebookentry.product2id, vlocity_cmt__Action__c, vlocity_cmt__AssetReferenceId__c, vlocity_cmt__AttributeMetadataChanges__c, vlocity_cmt__AttributeSelectedValues__c, vlocity_cmt__BillingAccountId__c, vlocity_cmt__CatalogItemReferenceDateTime__c, vlocity_cmt__CpqCardinalityMessage__c, vlocity_cmt__CpqMessageData__c, vlocity_cmt__CpqPricingMessage__c, vlocity_cmt__CurrencyPaymentMode__c, vlocity_cmt__EffectiveOneTimeTotal__c, vlocity_cmt__EffectiveRecurringTotal__c, vlocity_cmt__FirstVersionOrderItemId__c, vlocity_cmt__FulfilmentStatus__c, vlocity_cmt__InCartQuantityMap__c, vlocity_cmt__IsChangesAllowed__c, vlocity_cmt__ItemName__c, vlocity_cmt__LineNumber__c, vlocity_cmt__OneTimeCalculatedPrice__c, vlocity_cmt__OneTimeCharge__c, vlocity_cmt__OneTimeManualDiscount__c, vlocity_cmt__OneTimeTotal__c, vlocity_cmt__OrderGroupId__c, vlocity_cmt__ParentItemId__c, vlocity_cmt__PricingLogData__c, vlocity_cmt__Product2Id__c, vlocity_cmt__ProductGroupKey__c, vlocity_cmt__ProductHierarchyGroupKeyPath__c, vlocity_cmt__ProductHierarchyPath__c, vlocity_cmt__ProvisioningStatus__c, vlocity_cmt__ServiceAccountId__c, vlocity_cmt__SubAction__c, vlocity_cmt__SupersededOrderItemId__c, vlocity_cmt__SupplementalAction__c, vlocity_cmt__billingaccountid__r.id, vlocity_cmt__billingaccountid__r.name, vlocity_cmt__recurringcalculatedprice__c, vlocity_cmt__recurringcharge__c, vlocity_cmt__recurringdiscountprice__c, vlocity_cmt__recurringmanualdiscount__c, vlocity_cmt__recurringtotal__c, vlocity_cmt__rootitemid__c, vlocity_cmt__serviceaccountid__r.id, vlocity_cmt__serviceaccountid__r.name

    def test_get_cart_items_real(self):
        restClient.init('NOSDEV')
        res= CPQ.getCartItems_api('8013O000003keqvQAA')       
        print()

    def test_get_cart_items_real(self):
        restClient.init('NOSDEV')
        res = digitalCommerce.getOfferDetails('DC_CAT_MPO_CHILD_007','C_NOS_OFFER_017')
        print()


from . import restClient,utils,query,thread
import simplejson,sys
#------------------------------------------------------------------------------------------------------------------------------------------------
def dr(bundleName,inputData):
   """
   - bundleName: Name of the Data Raptor Bundle
   - inputData: input Data to the DR
   """
   action = f'/services/apexrest/{restClient.getNamespace()}/v2/DataRaptor'

   data = {
       "bundleName" : bundleName, 
       "objectList": inputData
   }
   call = restClient.callAPI(action, method="post", data=data)
   return call
#------------------------------------------------------------------------------------------------------------------------------------------------
def ipResponse(name,input,options=None,field=None):
   """
   Calls the IP and returns the response object
   - name: type_subtype
   - input: input data json.
   - options: the options json. 
   - field: if specified will return the value of the 'field' in response.  
   """
   res = ip(name,input,options)
   IPResult = res['IPResult']
   if 'response' in IPResult:
      IPResult = IPResult['response']

   if field != None and field in IPResult:
      return IPResult[field]
   return IPResult
#------------------------------------------------------------------------------------------------------------------------------------------------
def ip(name,input,options=None):
   """
   Calls the IP 
   - name: type_subtype
   - input: input data json.
   - options: the options json. 
   """
   action = f'/services/apexrest/vlocity_cmt/v1/GenericInvoke/vlocity_cmt.IntegrationProcedureService/{name}'
   data = {
      "sMethodName":name,
      "sClassName":"vlocity_cmt.IntegrationProcedureService",
      }

   if type(input) is dict:
      input = simplejson.dumps(input)

   data['input'] = input
   if options == None:
      options = "{}"
   else:
      if type(options) is dict:
         options = simplejson.dumps(options)      
   data['options'] =  options


   call = restClient.callAPI(action=action,data=data,method='post')
   lc = restClient.callSave('data1234',logRequest=True,logReply=False)
   utils.printJSON(call)
   #print(call)
   return call
#------------------------------------------------------------------------------------------------------------------------------------------------
def ip_response(call,field=None):
   """
   Get the response from the IP call response. 
   """
   IPResult = call['IPResult']
   if 'response' in IPResult:
      IPResult = IPResult['response']

   if field != None and field in IPResult:
      return IPResult[field]
   return IPResult
#------------------------------------------------------------------------------------------------------------------------------------------------
def get_IP_definitions():
   """
   Get the IP definitions. 
   Returns the procedure name, the IP name and the steps as strings. 
   """
   q = f"""select 
               Id,
               vlocity_cmt__Content__c,
               vlocity_cmt__OmniScriptId__r.name,
               vlocity_cmt__OmniScriptId__r.vlocity_cmt__ProcedureKey__c 
               from vlocity_cmt__OmniScriptDefinition__c 
               where vlocity_cmt__OmniScriptId__c in (select Id from vlocity_cmt__OmniScript__c where vlocity_cmt__OmniProcessType__c = 'Integration Procedure' and vlocity_cmt__IsActive__c = TRUE) """

   res = query.query(q)

   ip_definitions = []

   for record in res['records']:
      ip_definition = {
            'vlocity_cmt__ProcedureKey__c': record['vlocity_cmt__OmniScriptId__r']['vlocity_cmt__ProcedureKey__c'],
            'Name':                         record['vlocity_cmt__OmniScriptId__r']['Name'],
            'steps':                        []
      }
      ip_definitions.append(ip_definition)
      content = simplejson.loads(record['vlocity_cmt__Content__c'])
      for child in content['children']:
            ip_definition['steps'].append(child['name'])

   return ip_definitions

def call_chainable_ip(name,input,options=None):
   action = f'/services/apexrest/callipasynch/v1/{name}'
   action = f'/services/apexrest/callip/v1/{name}'
   action = f'/services/apexrest/callip/v2/{name}'
   action = f'/services/apexrest/callip/v3/{name}'

   data = {
      "sMethodName":name,
      "sClassName":"vlocity_cmt.IntegrationProcedureService",
      }

   data['input'] = input
   if options == None:
      options = {}
   data['options'] =  options


   call = restClient.callAPI(action=action,data=data,method='post')
   return call
#------------------------------------------------------------------------------------------------------------------------------------------------
def remoteClass(className,method,input,options=None):
   """
   Calls a remote APEX class inside Salesforce 
   - className: the callable APEX class name
   - method: the method to invoke.
   - input: the options json. 
   - options: the options json 
   """
   action = f'/services/apexrest/vlocity_cmt/v1/GenericInvoke/{className}/{method}'

   if type(input) is dict:
      input = simplejson.dumps(input)

   if options == None:
      options = {
         "postTransformBundle":"",
         "preTransformBundle":"",
         "useQueueableApexRemoting":False,
         "ignoreCache":False,
         "vlcClass":"B2BCmexAppHandler",
         "useContinuation":False
         }         
      
   data = {
      "sMethodName":method,
      "sClassName":className,
      "input":input,
      "options":simplejson.dumps(options)
      }

   call = restClient.callAPI(action=action,data=data,method='post')
   return call
#------------------------------------------------------------------------------------------------------------------------------------------------
def get_attachments(date1=None,date2=None,limit=None,Id=None):
   #date1 = "2023-01-01"
   #date2 = "2023-02-01"
   if date1 == None and limit == None and Id == None:
       return False
   
   if date1 != None:
      w2 = f" and LastModifiedDate<{date2}T00:00:00.00Z" if date2 != None else ''
      where = f" where vlocity_cmt__DRBundleName__c = 'None Specified' and vlocity_cmt__AsyncApexJobId__c = null and vlocity_cmt__GlobalKey__c != null and LastModifiedDate>{date1}T00:00:00.00Z {w2}"
   if limit !=None:
      where = f"  limit {limit}"
   if Id != None:
      where = f" where Id = '{Id}'"


   q0 = f"select Id,vlocity_cmt__GlobalKey__c from  vlocity_cmt__DRBulkData__c {where}"
   bulk_data_records = query.query(q0)

   if limit != None:
      idl = [r['Id'] for r in bulk_data_records['records']]
      q = f"select ID,Body,ParentId,LastModifiedDate from Attachment where ParentId in ({query.IN_clause(idl)})"
   else:
      q = f"select ID,Body,ParentId,LastModifiedDate from Attachment where ParentId in (select Id from  vlocity_cmt__DRBulkData__c {where} )"

   attachments = query.query_base64(q)

   for attachment in attachments['records']:
      attachment['vlocity_cmt__GlobalKey__c'] = [bdr['vlocity_cmt__GlobalKey__c'] for bdr in bulk_data_records['records'] if bdr['Id']==attachment['ParentId']][0]



   return attachments

def print_attachments(date1=None,date2=None,limit=None,Id=None,toFile=False):
   attachments = get_attachments(date1,date2,limit,Id)
   if attachments == False:
       return False
   ip_definitions = _get_IP_definitions()

   def do_work(attachment):
      attachment['log'] = restClient.requestWithConnection(action=attachment['Body'])
      return attachment
   
   def on_done(attachment,result):
      a=1

   thread.execute_threaded(attachments['records'],None,do_work,on_done,threads=num_threads)

   result = []

   for attachment in attachments['records']:
      if 'log' not in attachment:
          continue
      log = attachment['log']
      if 'vlcDebug' in log:
            possible_ips = _findMatch(log,ip_definitions)
            if len(possible_ips) >0:  attachment['possible'] = possible_ips
            else: attachment['best_match'] = _bestMach(log,ip_definitions)
            result.append(attachment)
            print("",end='.')


   found = [attachment for attachment in attachments['records'] if 'possible' in attachment]
   best_match = [attachment for attachment in attachments['records'] if 'best_match' in attachment]

   newlist = sorted(found, key=lambda d: d['LastModifiedDate']) 

   print_list = []
   for y,record in enumerate(newlist):
      for x,pos in enumerate(record['possible']):
            if len(pos['missing']) == 0:
               rec = {'name':"?","type":"?"}
            else:
               rec = pos['missing'][0]
            p = {
               'LastModifiedDate':record['LastModifiedDate'][0:19] if x==0 else "",
         #      'Id':record['Id'],
               'ParentId':record['ParentId'] if x==0 else "",
               'vlocity_cmt__GlobalKey__c':record['vlocity_cmt__GlobalKey__c'] if x==0 else "",
               'IP':pos['ip'],
               'debug':pos['executed'],
              # 'missing' :",".join([f"{rec['name']}-[{rec['type']}]" for rec in pos['missing']])
               'next Step':f"{rec['name']}-[{rec['type']}]" 

            }
            print_list.append(p)
   utils.printFormated(print_list)

   if Id != None:
      print()
      for step in record['possible'][0]['ip_steps']:
         if step in record['possible'][0]['debug_steps']:
            print(f"     {utils.CGREEN}{step}{utils.CEND}")
         else:
            print(f"     {step}")

      json_formatted_str = simplejson.dumps(attachment['log'], indent=2, ensure_ascii=False)
      if toFile == True:
         original_stdout = sys.stdout
         filename = f"{Id}.json"
         with open(filename, 'w') as f:
            sys.stdout = f 
            print(json_formatted_str)
            sys.stdout = original_stdout 
            print()
            print(f"   File {filename} created.")
      else:  
         print(json_formatted_str)

   return print_list

num_threads=15

def _get_IP_definitions():
   q = f"""select 
               Id,
               vlocity_cmt__Content__c,
               vlocity_cmt__OmniScriptId__r.name,
               vlocity_cmt__OmniScriptId__r.vlocity_cmt__ProcedureKey__c 
               from vlocity_cmt__OmniScriptDefinition__c 
               where vlocity_cmt__OmniScriptId__c in (select Id from vlocity_cmt__OmniScript__c where vlocity_cmt__OmniProcessType__c = 'Integration Procedure' and vlocity_cmt__IsActive__c = TRUE) """

   res = query.query(q)

   ip_definitions = []

   for record in res['records']:
      try:
         ip_definition = {
               'vlocity_cmt__ProcedureKey__c': record['vlocity_cmt__OmniScriptId__r']['vlocity_cmt__ProcedureKey__c'],
               'Name':                         record['vlocity_cmt__OmniScriptId__r']['Name'],
               'steps':                        [],
               'steps_names':[]
         }

         ip_definitions.append(ip_definition)
         content = simplejson.loads(record['vlocity_cmt__Content__c'])

         def getChildStep(child,parent_conditional='',parent_loop=False):
               if 'installCoverageEligibility' == child['name']:
                  a=1
               step = {
                  'name':child['name'],
                  'label':child['propSetMap']['label'] if 'label' in child['propSetMap'] else '',
                  'executionConditionalFormula':child['propSetMap']['executionConditionalFormula'] if 'executionConditionalFormula' in child['propSetMap'] else '',
                  'chainOnStep':child['propSetMap']['chainOnStep'] if 'chainOnStep' in child['propSetMap'] else False,
                  'indexInParent':child['indexInParent'],
                  'type':child['type'],
                  'loop_element':parent_loop
               }    
               if parent_conditional != '':
                  step['executionConditionalFormula'] = f"{step['executionConditionalFormula']} - {parent_conditional}"

               if child['type'] == 'Loop Block':
                  step['loop_element'] = True

               return step

         def addChildren(child,parent_conditional='',parent_loop=False):
               step = getChildStep(child,parent_conditional,parent_loop)
               ip_definition['steps'].append(step)
               ip_definition['steps_names'].append(step['name'])

               if 'children' in child and len(child['children'])>0:
                  for child1 in child['children']:
                     if 'eleArray' in child1:
                           for eleChild in child1['eleArray']:
                              parent_loop = True if child['type'] == 'Loop Block' or parent_loop == True  else False
                              addChildren(eleChild,parent_conditional=step['executionConditionalFormula'],parent_loop=parent_loop)
                     else:
                           a=1

         for child in content['children']:
               addChildren(child)
      except Exception as e:
          print(f"IP_Definition Error: {e}")
   return ip_definitions

def _findMatch(log,ip_definitions):
   possible_ips = []

   sequenceExecute = log['vlcDebug']['executionSequence']
   sequence = [key[0:-6] for key in log.keys() if key.endswith('Status') and type(log[key]) is bool]
   if sorted(sequence) == sorted(sequenceExecute):
      a=1
   else:
      a=1

   for ip_definition in ip_definitions:           
      if len(sequence) > len(ip_definition['steps_names']):  continue

      if ip_definition['vlocity_cmt__ProcedureKey__c'] == 'woo_getBundleNBOOffer':
            a=1          
      found = True
      for step in sequence:
            if step not in ip_definition['steps_names']: 
               found = False
               break
      if found: 
            posible = {
               'ip':ip_definition['vlocity_cmt__ProcedureKey__c'],
               'executed':"",
               'ip_steps':ip_definition['steps_names'],
               'debug_steps':sequence,
               'missing':[]
            }
            for ip_step in ip_definition['steps']:
               ex = "1" if ip_step['name'] in sequence or ip_step['name'] in sequenceExecute else "0"
               if '0' not in posible['executed']:
                  if ex == "0" and ip_step['executionConditionalFormula'] != "": 
                        ex = '_'
                  if ex == "0" and ip_step['loop_element'] == True: ex = 'L'
               if ex == '0': posible['missing'].append(ip_step)

               posible['executed'] = f"{posible['executed']}{ex}"
            if '1' not in posible['executed']: continue
            if posible['executed'][0]=='0': 
               continue
            if '01' in posible['executed']:
               continue
            possible_ips.append(posible)

   return possible_ips

def _bestMach(log,ip_definitions):
   possible_ips = []

   sequenceExecute = log['vlcDebug']['executionSequence']
   sequence = [key[0:-6] for key in log.keys() if key.endswith('Status') and type(log[key]) is bool]

   for ip_definition in ip_definitions:
      if ip_definition['vlocity_cmt__ProcedureKey__c'] == 'iUse_submitResumeOrder':
            a=1
      score = 0

      executed = ""
      for ip_step in ip_definition['steps']:
            ex = "1" if ip_step['name'] in sequence else "0"
            if ex == "0" and ip_step['executionConditionalFormula'] != "": ex = '_'
            if ex == "0" and ip_step['loop_element'] == True: ex = 'L' 
            executed = f"{executed}{ex}"

      missed = []
      for seq in sequence:
            if seq not in ip_definition['steps_names']:
               executed = f"{executed}*"
               missed.append(seq)
      score = executed.count('1')

      if score>0:
            ip = {
               'name':ip_definition['vlocity_cmt__ProcedureKey__c'],
               'score':score,
               'ip_steps':len(ip_definition['steps']),
               'size':len(sequence),
               'execured':executed,
               'ip_stesps':ip_definition['steps'],
               'debug_steps':sequence,
               'missed':missed
            }
            possible_ips.append(ip)
   
   possible_ips_sorted = sorted(possible_ips, key=lambda d: d['score'])
   possible_ips_sorted.reverse()

   return possible_ips_sorted




import unittest,simplejson
from InCli.SFAPI import restClient,query,Sobjects

class Test_Utilities(unittest.TestCase):
    def test_limits(self):
        restClient.init('DEVNOSCAT2')
        action = '/services/data/v51.0/limits'
        res = restClient.callAPI(action)

        print()


    def test_id(self):
        restClient.init('DEVNOSCAT2')
        action = '/services/data/v51.0'
        res = restClient.callAPI(action)

        print()

    def test_id(self):
        restClient.init('DEVNOSCAT2')
        action = '/services/data/v51.0'
        res = restClient.callAPI(action)
        for key in res.keys():
            print()
            action = res[key]
            res1 = restClient.callAPI(action)
            print(action)
            print(res1)

        print()
    
    def test_select(self):
        restClient.init('DEVNOSCAT4')

        q = f"select fields(all) from vlocity_cmt__VlocityTrackingEntry__c order by vlocity_cmt__Timestamp__c desc limit 100"
        res = query.query(q)
        for r in res['records']:
            ll = simplejson.loads(r['vlocity_cmt__Data__c'])
            json_formatted_str = simplejson.dumps(ll, indent=2, ensure_ascii=False)
            print(json_formatted_str)
            print()
    def test_delete_logs(self):
        restClient.init('demo240')

        q = "select Id from ApexLog where LogUserId='005Dm000000Yj0rIAC' "
        res = query.query(q)

        id_list = [record['Id'] for record in res['records']]
        
        Sobjects.deleteMultiple('ApexLog',id_list)
        print()

    def delete(self,q):
        res = query.query(q)

        id_list = [record['Id'] for record in res['records']]
        
        Sobjects.deleteMultiple('ApexLog',id_list)
        print()

    def test_delete_something(self):
        restClient.init('DEVNOSCAT4')

        q = "select Id from vlocity_cmt__ConfigurationSnapshot__c  "
        self.delete(q)


from . import restClient,utils


class TimeStats:
    time_records = []
    time_record = {}

    def new(self,fields=None):
        self.time_record = {}
        self.time_records.append(self.time_record)
        if fields == None: return
        for field in fields:
            self.time_record[field] = ''
    def time(self,field):
        self.time_record[field] = restClient.getLastCallTime()
    
    def time_no(self,field,value):
        self.time_record[field] = value
    def print(self):
        utils.printFormated(self.time_records)





from . import restClient, objectUtil,digitalCommerce,utils,thread,query,CPQ
import time
import sys

def _getItemAtPath(obj,path,field='ProductCode'):
    for p in path.split(':'):
        obj = objectUtil.getSibling(obj,field,p)  
    return obj

def updateOfferField(offerDetails,path,field,value,pathField='ProductCode'):
    obj = _getItemAtPath(offerDetails,path,pathField)
    if obj == '':
        raise ValueError(f'Object does not contain element by path {path}')
    if field in obj:
        obj[field] = value
    else:
        raise ValueError(f'Object does not contain field {field} by path {path}')

    return offerDetails

def updateOfferAttribute(offerDetails,path,AttributeCategory,productAttribute,value):
    obj = _getItemAtPath(offerDetails,path)
    obj = _getItemAtPath(obj,AttributeCategory,field='Code__c')
    obj = _getItemAtPath(obj,productAttribute,field='code')

    obj['userValues'] = value
 
    return offerDetails

def getBasketProductAttributes(basket,path):
    obj = _getItemAtPath(basket,path)
    obj = obj['attributeCategories']

    return obj

def getAction(basket,path,method='addChildBasketAction'):

    lpath = path.split(':')
    xpath = ":".join(lpath[:-1])
    offer = lpath[-1]


    obj = _getItemAtPath(basket,xpath)
    action = _getItemAtPath(obj,method,field='method')

    action['params']['offer'] = offer
    action['link'] = f"/services/apexrest/{restClient.getNamespace()}{action['link']}"



    return action

def printCatalogs(catcatalogueCode=None):
    catalogs = digitalCommerce.getCatalogs(catcatalogueCode)
    print()
    print(f'Catalogs in the Org {utils.CGREEN+ restClient._currentConnectionName + utils.CEND}:')
    utils.printFormated(catalogs,"Name:vlocity_cmt__CatalogCode__c:vlocity_cmt__IsActive__c:vlocity_cmt__Description__c",rename="vlocity_cmt__IsActive__c%Active")
    print() 
    return catalogs

def checkOffers(path=None,quantity=None,catcatalogueCode=None,account=None,basketOps=False):

    catalogs = printCatalogs(catcatalogueCode)

    catsOffers = []
    numOffersList = []
    print()

    print(utils.CLIGHT_CYAN+"GetOffers per catalog."+utils.CEND)
    getOfferTimes = []
    def getOffersPerCatalog(catalog):

        if catalog['vlocity_cmt__IsActive__c'] == False:
            return 
        try:
            times = {
                'name':catalog['vlocity_cmt__CatalogCode__c'],
                'Error':'',
                "__color__":utils.CEND
            }
            offers = digitalCommerce.getOfferByCatalogue(catalog['vlocity_cmt__CatalogCode__c'],pagesize=400)
            times['elapsed'] = restClient.getLastCallTime()
            l = len(offers) if offers is not None else 0
            numOffersList.append(l)
            times['# Offers'] = l

        except Exception as e:
            times['elapsed'] = restClient.getLastCallTime()
            times['Error'] = e.args[0]['error']
            times['__color__'] = utils.CRED
            getOfferTimes.append(times)
            return   

        if offers == None:
            times['Error'] = "Has no offers."
            times['__color__'] = utils.CYELLOW
            getOfferTimes.append(times)
            return 
       #             obj['off effDate'] = offer[0]['vlocity_cmt__EffectiveStartDate__c'] if obj['type'] == 'Promotion' else offer[0]['vlocity_cmt__Product2Id__r'] 


        relationships = query.queryRecords(f"""select vlocity_cmt__SequenceNumber__c,
                                                        vlocity_cmt__EffectiveDate__c,
                                                        Name,
                                                        Id,
                                                        vlocity_cmt__Product2Id__r.Name,
                                                        vlocity_cmt__Product2Id__r.vlocity_cmt__EffectiveDate__c,
                                                        vlocity_cmt__Product2Id__r.ProductCode,
                                                        vlocity_cmt__PromotionId__r.Name,
                                                        vlocity_cmt__PromotionId__r.vlocity_cmt__EffectiveStartDate__c,
                                                        vlocity_cmt__PromotionId__r.vlocity_cmt__Code__c,
                                                        vlocity_cmt__IsActive__c,
                                                        vlocity_cmt__ItemType__c,
                                                        IsDeleted 
                                                        from vlocity_cmt__CatalogProductRelationship__c 
                                                        where vlocity_cmt__CatalogId__c = '{catalog['Id']}' order by vlocity_cmt__SequenceNumber__c limit 200""")

        catOffers = {
            'catCode':catalog['vlocity_cmt__CatalogCode__c'],
            'catId':catalog['Id'],
            'offers':offers,
            'relationships':relationships
        }
        catsOffers.append(catOffers)
        getOfferTimes.append(times)

    thread.processList(getOffersPerCatalog,catalogs,1)
    utils.printFormated(getOfferTimes)
    totalOffers = sum(numOffersList)

    for catOffers in catsOffers:
        print()
        print()

        print(utils.CLIGHT_CYAN+ f"Product relationships and offers in the API for catalog {catOffers['catCode']} "+utils.CEND)
        objs = []
        for relationship in catOffers['relationships']:
          #  if relationship['vlocity_cmt__ItemType__c'] == 'Promotion':
                fields = ['Name','Id','vlocity_cmt__ItemType__c','vlocity_cmt__IsActive__c','vlocity_cmt__Product2Id__c','vlocity_cmt__PromotionId__c']
                obj = {
                    "seq":relationship['vlocity_cmt__SequenceNumber__c'],
                    "effDate":relationship['vlocity_cmt__EffectiveDate__c'],
                    "Name_Relationship":relationship['Name'],
                    "Id_Relationship":relationship['Id'],
                    "type":relationship['vlocity_cmt__ItemType__c'],
                    "Actv":relationship['vlocity_cmt__IsActive__c'],
                    "Name_prod":relationship['vlocity_cmt__Product2Id__r']['Name'] if relationship['vlocity_cmt__Product2Id__r']!=None else '',
                    "Name_promo":relationship['vlocity_cmt__PromotionId__r']['Name'] if relationship['vlocity_cmt__PromotionId__r']!=None else '',
                    "P_effDate":relationship['vlocity_cmt__Product2Id__r']['vlocity_cmt__EffectiveDate__c'] if relationship['vlocity_cmt__Product2Id__r']!=None else relationship['vlocity_cmt__PromotionId__r']['vlocity_cmt__EffectiveStartDate__c'],
                    "P_Code":relationship['vlocity_cmt__Product2Id__r']['ProductCode'] if relationship['vlocity_cmt__Product2Id__r']!=None else relationship['vlocity_cmt__PromotionId__r']['vlocity_cmt__Code__c']
                }
                if 'T' in obj["P_effDate"]:
                    obj["P_effDate"] = obj["P_effDate"].split('T')[0]

                if obj['type'] == 'Promotion':
                    offer = [ offer for offer in catOffers['offers'] if offer['offerType'] == 'Promotion' and offer['Name']==obj['Name_promo']]
                if obj['type'] == 'Product':
                    offer = [ offer for offer in catOffers['offers'] if offer['offerType'] == 'Product' and offer['Name']==obj['Name_prod']]
                if len(offer)>0:
                    obj['Offer Name'] = offer[0]['Name']
                    obj['Offer Code'] = offer[0]['vlocity_cmt__Code__c'] if obj['type'] == 'Promotion' else offer[0]['ProductCode']
                    obj['P_isActive'] = offer[0]['vlocity_cmt__IsActive__c'] if obj['type'] == 'Promotion' else offer[0]['IsActive']
                    obj['isOrderable'] = offer[0]['vlocity_cmt__IsOrderable__c']
                    obj['__color__'] = utils.CEND
                else:
                    obj['NameP'] = 'No Offer'
                    obj['Code'] = ""
                    obj['PActv'] = ""
                    obj['isOrderable'] = ""     
                    obj['__color__'] = utils.CRED if obj['Actv'] == True else utils.CFAINT 

                obj['Name_promo'] = f"Promo: {obj['Name_promo']}" if obj['Name_promo']!='' else f"Prod: {obj['Name_prod']}"

                objs.append(obj)
        utils.printFormated(objs,exclude="Id_Relationship:type:Name_prod")

    print()
    print()

    if basketOps == True:
        print(utils.CLIGHT_CYAN+f"Getting offerDetails, createBasket, createBasket with config per offer. total offers {totalOffers}"+utils.CEND)
    else:        
        print(utils.CLIGHT_CYAN+f"Getting offerDetails. total offers {totalOffers}"+utils.CEND)


    offersList = []

    accountId = None

    if account != None:
        chunks = account.split(':')
        id = query.queryField(f" select Id from Account where {chunks[0]} = '{chunks[1]}'")
        if id==None:
            print(utils.CLIGHT_RED+ f"Cannot find an account where {chunks[0]} = '{chunks[1]}'"+utils.CEND)
        else:
            accountId=id

    for catOffers in catsOffers:
        for offerCount,offer in enumerate(catOffers['offers']):
            _offer = {
                'catalogCode':catOffers['catCode'],
                'offerCode':digitalCommerce.getOfferCode(offer),
                'basketOps':basketOps
            }
            if accountId != None:
                _offer['accountId']=id
            offersList.append(_offer)

    thread.processList(dba,offersList,50)
    newlist = sorted(_theTimes, key=lambda d: f"{d['catalog']}{d['offerCode']}")
    utils.printFormated(newlist)
    for error in _errors:
        utils.printException(error)

    print()

_theTimes = []
_errors = []
def dba(offerCodes):
    try:
        offerCode = offerCodes['offerCode']
        catalogCode = offerCodes['catalogCode']
        times = {
            'catalog':catalogCode,
            'offerCode':offerCode,
            'details':'',
            'createBasket':'',
            'afterConfig':'',
            '__color__':utils.CEND,
            "error":''
        }

        details = digitalCommerce.getOfferDetails(catalogCode,offerCode)
        times['details'] = restClient.getLastCallTime()

        if offerCodes['basketOps'] == True:
            basket = digitalCommerce.createBasket(catalogCode,offerCode)
            times['createBasket'] = restClient.getLastCallTime()

            if details['result']['offerDetails']['offer']['offerType'] == 'Promotion':
                try:
                    offerDetails = updateOfferField(details,'ATT_NOS_OTT_SUBSCRIPTION_ID','values','xxxxxx','code')
                except Exception as e:
                    if e.args[0] == 'Object does not contain element by path ATT_NOS_OTT_SUBSCRIPTION_ID':
                        offerDetails = updateOfferField(details,'ATT_NOS_PRICE_CONDITIONS','values','PRICE_COND_003','code')
                    else:
                        raise
            else:
                offerDetails = updateOfferField(details,'0001','Quantity',4,'lineNumber')

            basket2 = digitalCommerce.createBasketAfterConfig(catalogCode,offerDetails)
            times['afterConfig'] = restClient.getLastCallTime()
        if 'accountId' in offerCodes and 1==2:
            times['cart']=''
            cart = digitalCommerce.createCart(offerCodes['accountId'],catalogCode,basket2['cartContextKey'],createAsset=False)
            times['cart']=restClient.getLastCallTime()

            times['delete']=''
            delete = CPQ.deleteCart(cart['orderId'])
            times['delete']=restClient.getLastCallTime()

            print()

    except Exception as e:
        if len(e.args)> 0  and 'error' in e.args[0]:
            times['error']=e.args[0]['errorCode']
        else:
            times['error']=f"{e}"

        _errors.append(e)

        times['__color__'] = utils.CLIGHT_RED

    _theTimes.append(times)




import unittest,simplejson,sys
from InCli.SFAPI import restClient,query,Sobjects,utils,jsonFile,file_csv


class Test_Objects(unittest.TestCase):

    def test_objects(self):
        restClient.init('NOSQSM')
        
        res = Sobjects.get_with_only_id('a3m3O000000KCjCQAW')

        print()     
    def test_limits(self):
        restClient.init('DEVNOSCAT2')
        action = '/services/data/v51.0/limits'
        res = restClient.callAPI(action)

        print()


    def test_id(self):
        restClient.init('DEVNOSCAT2')
        action = '/services/data/v51.0'
        res = restClient.callAPI(action)

        print()

    def test_id(self):
        restClient.init('DEVNOSCAT2')
        action = '/services/data/v51.0'
        res = restClient.callAPI(action)
        for key in res.keys():
            print()
            action = res[key]
            res1 = restClient.callAPI(action)
            print(action)
            print(res1)

        print()

    def getAttributes(self,product):
        attr_str = product['vlocity_cmt__AttributeMetadata__c']
        if attr_str == None: return None
        atributes = simplejson.loads(attr_str)

        atts = []
        for atribute in atributes['records']:
            for productAttribute in atribute['productAttributes']['records']:
                if 'values' not in productAttribute:
                    a=1
                if productAttribute['values'] == None:
                    a=1
                att = {
                    'att':atribute['Code__c'],
                    'pAtt':productAttribute['code'],
                    'type':productAttribute['inputType'],
                    'len':len(productAttribute['values']) if 'values' in productAttribute and productAttribute['values'] != None else 0
                }

                atts.append(att)
        return atts

    def getChildProducts(self,product):
        children = []
        childItems = query.queryRecords(f"select fields(all) from vlocity_cmt__ProductChildItem__c where vlocity_cmt__ParentProductId__c='{product['Id']}' and vlocity_cmt__IsOverride__c = False limit 200")
        if len(childItems) == 0:
            return []

        for childItem in childItems:
            if childItem['vlocity_cmt__ChildProductId__c'] == None:
                continue
            prod = Sobjects.getF('Product2',f"Id:{childItem['vlocity_cmt__ChildProductId__c']}")['records'][0]
            print(prod['Name'])

            child = {
                'Name':childItem['vlocity_cmt__ChildProductName__c'],
                'virtual':childItem['vlocity_cmt__IsVirtualItem__c'],
                'Id':childItem['vlocity_cmt__ChildProductId__c'],
                'attributes':self.getAttributes(prod),
                'children':self.getChildProducts(prod),
                'mmq':f"({childItem['vlocity_cmt__MinMaxDefaultQty__c']})".replace(' ',''),
                "child_mm":f"[{int(childItem['vlocity_cmt__MinimumChildItemQuantity__c'])},{int(childItem['vlocity_cmt__MaximumChildItemQuantity__c'])}]"
            }
            children.append(child)
            print(childItem['vlocity_cmt__ChildProductName__c'])

        return children

    code = 'PROMO_NOS_OFFER_005'

    def test_getF(self):
        restClient.init('NOSDEV')

        res = Sobjects.getF('Product2',"ProductCode:C_NOS_OFFER_001")

        code='PROMO_WOO_FIXED_INTERNET_MOBILE_12_MONTHS_008'
        promo = Sobjects.getF('vlocity_cmt__Promotion__c',f"vlocity_cmt__Code__c:{self.code}")

        promoItems = query.queryRecords(f"select fields(all) from vlocity_cmt__PromotionItem__c where vlocity_cmt__PromotionId__c='{promo['records'][0]['Id']}' limit 200")
        root = {
            'children':[],
            'Name':'root',
            'Id':'NA',
            'attributes':"NA"
        }
        for promoItem in promoItems:
            prods = Sobjects.getF('Product2',f"Id:{promoItem['vlocity_cmt__ProductId__c']}")

            for prod in prods['records']:
                _product = {
                    'Name':prod['Name'],
                    'virtual':False,
                    'Id':prod['Id'],
                    'attributes':self.getAttributes(prod),
                    'children':self.getChildProducts(prod),
                    'mmq':"",
                    'child_mm':""
                }
                root['children'].append(_product)
        jsonFile.write(f'prod123_{self.code}',root)

        self.printProduct(root)

    def test_print_from_file(self):
        root = jsonFile.read(f'prod123_{self.code}')
        self.printProduct(root)


    def get_product_decomposition(self,product):
        q = f"select fields(all) from vlocity_cmt__DecompositionRelationship__c where vlocity_cmt__SourceProductId__c='{product['Id']}' limit 200"
        res = query.query(q)
        product['decompositions'] = []
        for record in res['records']:
            mappring_data_str = record['vlocity_cmt__MappingsData__c']
            mappring_data = simplejson.loads(mappring_data_str) if mappring_data_str != None else None

            mappings =[]
            if mappring_data != None:
                for mapping in mappring_data:
                    if mapping['mapping_type'] == 'ad-verbatim':
                        if mapping['source_type'] == 'Attribute':
                            mappings.append({
                                'from':mapping['source_attr_code'],
                                'to':mapping['destination_attr_code']
                            })
                        elif mapping['source_type'] == 'Field':
                            mappings.append({
                                'from':mapping['source_field_name'],
                                'to':mapping['destination_attr_code']
                            })
                        else:
                            a=1
                    elif mapping['mapping_type'] == 'static':
                        mappings.append({
                            'from':'Static',
                            'to':mapping['destination_attr_code']
                        })

                    elif mapping['mapping_type'] == 'list':
                        if mapping['source_type'] == 'Attribute':
                            mappings.append({
                                'from':mapping['source_attr_code'],
                                'to':mapping['destination_attr_code']
                            })
                        elif mapping['source_type'] == 'Field':
                            mappings.append({
                                'from':mapping['source_field_name'],
                                'to':mapping['destination_attr_code']
                            })
                        else:
                            a=1
                    else:
                        a=1

            condition_data_str = record['vlocity_cmt__ConditionData__c']
            condition_data = simplejson.loads(condition_data_str) if condition_data_str!=None else None

            decomposition = {
                'conditions' :len(condition_data['singleconditions']) if condition_data != None else 0,
                'mapping_rules':len(mappring_data) if mappring_data != None else 0,
                'Name':record['Name'],
                'destination_Id':record['vlocity_cmt__DestinationProductId__c'],
                'mappings':mappings
            }
            product['decompositions'].append(decomposition)

            if decomposition['destination_Id'] != None:

                q = f"select fields(all) from vlocity_cmt__DecompositionRelationship__c where vlocity_cmt__SourceProductId__c='{decomposition['destination_Id']}' limit 200"
                res = query.query(q)
                if len(res['records'])>0:
                    level = 2 if 'level' not in product else product['level']+1
                    if 'level' in product:
                        a=1
                    fake_product = {
                        'Id':decomposition['destination_Id'],
                        'level':level
                    }
                    self.get_product_decomposition(fake_product)
                    decomposition['next_level'] = fake_product['decompositions']
                    print()
            else:
                a=1

            print(record['Name'])

        print()

    def test_decomposition_rules(self):
        restClient.init('NOSDEV')

        root = jsonFile.read(f'prod123_{self.code}')

        def get_decomposition(product):
            self.get_product_decomposition(product)
            for child in product['children']:
                get_decomposition(child)

        for children in root['children']:
            get_decomposition(children)

        jsonFile.write(f'prod123_decomposed_{self.code}',root)


        print()

    def flatten(self,product):
        a=1




    def printProduct(self,products,path=[]):

        filename = f'prod123_decomposed_{self.code}_csv.csv'
        original_stdout = sys.stdout

        with open(filename, 'w') as f:
            sys.stdout = f 

            def printProduct_inner(products,path=[]):
                if products == None:
                    a=1
                for product in products['children']:
                    spath = path.copy()
                    spath.append(f"{product['Name']}{product['mmq']}{product['child_mm']}")

                    self.printAttribute(spath,product['attributes'])
                    self.print_decomposition(spath,product['decompositions'])
                    try:
                        printProduct_inner(product,spath)
                    except Exception as e:
                        print(e)
            printProduct_inner(products)
        sys.stdout = original_stdout 

        print()


    def printAttribute(self,path,attributes):
        spath = path.copy()
        while len(spath)<5:
            spath.append("")

        _path = ";".join(spath)

        if attributes == None: 
     #       print(f"{_path}")
            return
        for atttribute in attributes:
            print(f"{_path};{atttribute['att']}-{atttribute['pAtt']};{atttribute['type']};{atttribute['len']}")


    def print_decomposition(self,path,decompostions):
        spath = path.copy()
        while len(spath)<8:
            spath.append("")

        _path = ";".join(spath)

        if decompostions == None: 
     #       print(f"{_path}")
            return

        def print_next_level(path,next_decompositions,level):
            for next_decomposition in next_decompositions:
                xx = []
                while len(xx)< (3 + (level-1)):
                    xx.append("")
                spaces = ";".join(xx)
                for next_mapping in next_decomposition['mappings']:
                    print(f"{path};{spaces};{next_decomposition['Name']};{next_mapping['from']}->{next_mapping['to']}")
                if 'next_level' in next_decomposition:
                    print_next_level(f"{_path};{decomposition['Name']};{next_decomposition['Name']}",next_decomposition['next_level'],level+1)

        for decomposition in decompostions:
            for mapping in decomposition['mappings']:
                print(f"{_path};{decomposition['Name']};{decomposition['conditions']};{decomposition['mapping_rules']};{mapping['from']}->{mapping['to']}")
            if 'next_level' in decomposition:
                print_next_level(f"{_path};{decomposition['Name']}",decomposition['next_level'],1)

    def test_print_from_file_deco(self):
        root = jsonFile.read(f'prod123_decomposed_{self.code}')
        self.printProduct(root)

#####################################################################
    def printProduct2(self,products,path=[]):

        filename = f'prod123_decomposed2_{self.code}_csv.csv'
        original_stdout = sys.stdout

        with open(filename, 'w') as f:
            sys.stdout = f 

            def printProduct_inner(products,path=[]):
                if products == None:
                    a=1
                for product in products['children']:
                    spath = path.copy()
                    virtual = " VIRTUAL " if product['virtual'] == True else ""
                    spath.append(f"{virtual}{product['Name']}{product['mmq']}{product['child_mm']}")

                    self.printAttribute2(spath,product['attributes'])
                    self.print_decomposition2(spath,product['decompositions'])
                    try:
                        printProduct_inner(product,spath)
                    except Exception as e:
                        print(e)
            printProduct_inner(products)
        sys.stdout = original_stdout 

        print()


    def printAttribute2(self,path,attributes):
        spath = path.copy()
        while len(spath)<5:
            spath.append("")

        _path = ";".join(spath)

        if attributes == None: 
     #       print(f"{_path}")
            return
        for atttribute in attributes:
            att = f"{atttribute['att']}-{atttribute['pAtt']}"
            print(f"{_path};AT:  {att};AT:{atttribute['type']} {atttribute['len']}")


    def print_decomposition2(self,path,decompostions):
        spath = path.copy()
        while len(spath)<5:
            spath.append("")

        _path = ";".join(spath)

        if decompostions == None: 
     #       print(f"{_path}")
            return

        def print_next_level(path,next_decompositions,level):
            for next_decomposition in next_decompositions:
                xx = []
                while len(xx)< (3 + (level-1)):
                    xx.append("")
                spaces = ";".join(xx)
                spaces = ''
                for next_mapping in next_decomposition['mappings']:
                    print(f"{path};DE: {next_decomposition['Name']};MAP: {next_mapping['from']}->{next_mapping['to']}")
                if 'next_level' in next_decomposition:
                    print_next_level(f"{_path};DE: {decomposition['Name']};DE: {next_decomposition['Name']}",next_decomposition['next_level'],level+1)

        for decomposition in decompostions:
            for mapping in decomposition['mappings']:
                print(f"{_path};DE: {decomposition['Name']}  C:{decomposition['conditions']};MAP: {mapping['from']}->{mapping['to']}")
            if 'next_level' in decomposition:
                print_next_level(f"{_path};DE: {decomposition['Name']}  C:{decomposition['conditions']}",decomposition['next_level'],1)

    def test_print_from_file_deco2(self):
        root = jsonFile.read(f'prod123_decomposed_{self.code}')
        self.printProduct2(root)